"""Main module."""

from concurrent.futures import ThreadPoolExecutor
import requests
import logging
import os
import configparser
import json
from pathlib import Path
from pathlib import PurePath
import re
import base64
import hashlib
from io import StringIO, IOBase
from typing import Dict, List, Tuple


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

VALID_STRUCTURE_SEARCH_TYPES = {"substructure", "duplicate",
                        "duplicate_tautomer", "duplicate_no_tautomer",
                        "stereo_ignore", "full_tautomer", "substructure",
                        "similarity", "full"}

def isBase64(s):
    """Checks if a string is base64 encoded.
    """
    if not isinstance(s, str):
        return False
    return (len(s) % 4 == 0) and re.match('^[A-Za-z0-9+/]+[=]{0,2}$', s)


def creds_from_file(fpath, profile="default"):
    """Fetches crecentials from a file

    Retrieves credentials from a file.

    Args:
        fpath: Path to the credentials file
        profile: Optional string which specifies which section to read from
        the credentials file.  Defaults to "default"
    Returns:
        A section of the configuration file.

    """
    config = configparser.ConfigParser()
    config.read(fpath)
    return config[profile]


def get_default_credentials(profile="default"):
    """Fetches default credentials to use for client authentication.

    Retrievies credentials from ~/.acas/credentials file or
    preferentially from the environment variables
    ACAS_API_{USERNAME, PASSWORD and URL}

    Args:
        profile: Optinal string which specifies which section to read from
        the credentials file.
    Returns:
        A dict of credentials fetched. For example:

        .. code-block:: python

            {'username': "USERNAME",
            'password': "PASSWORD",
            'url': "URL"}

        Example setting via environment variables:

        .. code-block:: console

                $ export ACAS_API_USERNAME=bob
                $ export ACAS_API_PASSWORD=secret
                $ export ACAS_API_URL=http://localhost:3000

        If any one of the three environment variables are not found then the
        credentials file at ``~/.acas/credentials`` is used.

        Example credentials file:

        .. code-block:: ini

            [default]
            username=bob
            password=secret
            url=http://localhost:3000

        If the environment variable ``ACAS_API_PROFILE`` is set then it will be
        used as the default credential section from the credentials file
        otherwise, the ``[default]`` section is used. For example:

        .. code-block:: console

                $ export ACAS_API_PROFILE=myserver

        The ``myserver`` section credentials will be read:

        .. code-block:: ini

            [mysever]
            username=bob
            password=secret
            url=http://localhost:3000

    """
    try:
        data = {}
        data['username'] = os.environ['ACAS_API_USERNAME']
        data['password'] = os.environ['ACAS_API_PASSWORD']
        data['url'] = os.environ['ACAS_API_URL']
        logger.debug("Using credentials from ACAS_API_CREDENTIALS environment"
                     " variable.")
    except KeyError:
        homedir = os.environ['HOME']
        creds_file = Path(homedir, '.acas', 'credentials')
        logger.debug("Looking for credentials in {}".format(creds_file))
        if 'ACAS_API_PROFILE' in os.environ:
            profile = os.environ['ACAS_API_PROFILE']
        else:
            profile = profile
        data = creds_from_file(creds_file, profile)
    return {'username': data['username'], 'password': data['password'],
            'url': data['url']}


def get_entity_value_by_state_type_kind_value_type_kind(entity, state_type,
                                                        state_kind, value_type,
                                                        value_kind):
    """Get a value from an acas entity dict object.

    Gets a specific value from an acas entity dict object.

    Args:
        entity: Any ACAS entity (protocol, experiment, analysis_group,
        treatment_group, container...etc.)
        state_type: String. The state type of the value.
        state_kind: String. The state kind of the value.
        value_type: String. The value type of the value.
        value_kind: String. The value type of the value.

    Returns:
        A dict object representing the value if it exits. Otherwise it
        returns None

    """
    value = None
    if len(entity["lsStates"]) > 0:
        for s in entity["lsStates"]:
            if (not s["deleted"] and not s["ignored"] and
                    s["lsType"] == state_type and s["lsKind"] == state_kind):
                for v in s["lsValues"]:
                    if (not v["deleted"] and not v["ignored"] and
                            v["lsType"] == value_type and
                            v["lsKind"] == value_kind):
                        value = v
                        break
    return value

def get_entity_values_by_state_type_kind_value_type(entity, state_type,
                                                        state_kind, value_type):
    """Gets values from an acas entity dict object.

    Gets values from an acas entity dict object by state type, state kind and value type.

    Args:
        entity: Any ACAS entity (protocol, experiment, analysis_group,
        treatment_group, container...etc.)
        state_type: String. The state type of the value.
        state_kind: String. The state kind of the value.
        value_type: String. The value type of the value.

    Returns:
        A list of dict objects representing the values if they exits. Otherwise it
        returns an empty list

    """
    values = []
    if len(entity["lsStates"]) > 0:
        for s in entity["lsStates"]:
            if (not s["deleted"] and not s["ignored"] and
                    s["lsType"] == state_type and s["lsKind"] == state_kind):
                for v in s["lsValues"]:
                    if (not v["deleted"] and not v["ignored"] and
                            v["lsType"] == value_type):
                        values.append(v)
    return values

def get_entity_label_by_label_type_kind(entity: dict, label_type: str, label_kind: str) -> str:
    """Get a label from an acas entity dict object.

    Gets a specific label from an acas entity dict object.

    Args:
        entity: Any ACAS entity (protocol, experiment, analysis_group,
        treatment_group, container...etc.)
        label_type: String. The label type of the value.
        label_kind: String. The label kind of the value.

    Returns:
        A dict object representing the label if it exits. Otherwise it
        returns None

    """
    label = None
    for l in entity["lsLabels"]:
        if (not l["deleted"] and not l["ignored"] and
                l["lsType"] == label_type and l["lsKind"] == label_kind):
            label = l
            break
    return label

def sdf_iterator(iteratable):
    data = []
    for line in iteratable:
        data.append(line)
        if line.startswith("$$$$"):
            yield "".join(data)
            data = []

def get_mol_as_dict(mol):
    """
    Returns a dict representation of a molecule in cluding the mol block, the ctab and the properties as a key value pair
    """
    lines = mol.split("\n")
    properties = {}
    property = None
    ctab = None
    ctab_complete = False
    for line in lines:
        if line.startswith(">"):
            ctab_complete = True
            if property is not None:
                properties[property] = properties[property] = "\n".join(properties[property])
            property = line.split("<")[1].split(">")[0].strip()
            properties[property] = []
        else:
            if not ctab_complete:
                if ctab is None:
                    ctab = line + "\n"
                else:
                    ctab += line + "\n"
            if property is not None:
                if line.strip() != "":
                    properties[property].append(line.strip())
                else:
                    properties[property] = "\n".join(properties[property])
                    property = None
    return {"mol": mol, "ctab": ctab, "properties": properties}

def parse_file(file_content, file_extension):
    """Parse content from a string into an extension specific format

    Args:
        file_content (str): Content of the file
        file_extension (str): Extension of the file

    Returns:
        Parsed content of the file in a format specific to the extension
    """
    if file_extension == '.sdf':
        return parse_sdf(file_content)
    elif file_extension == '.json':
        return json.loads(file_content)
    else:
        return None

def parse_sdf(file_content):
    """Parse an SDF file content

    Parse an SDF file

    Args:
        file_content (bytes): Content of the file

    Returns:
        Parsed content of the file
    """
    sdf_data = []
    file_content = StringIO(file_content.decode('utf-8'))
    for e, mol in enumerate(sdf_iterator(file_content)):
        mol_dict = get_mol_as_dict(mol)
        sdf_data.append(mol_dict)
    return sdf_data 


class client():

    def __init__(self, creds):
        self.username = creds['username']
        self.password = creds['password']
        self.url = creds['url']
        self.session = self.getSession()

    def close(self):
        self.session.close()

    def getSession(self):
        data = {
            'username': self.username,
            'password': self.password
        }
        session = requests.Session()
        resp = session.post("{}/login".format(self.url),
                            headers={'Content-Type': 'application/json'},
                            data=json.dumps(data),
                            allow_redirects=False)
        resp.raise_for_status()
        if resp.status_code == 302 and 'location' in resp.headers and resp.headers.get('location') == "/login":
            raise RuntimeError("Failed to login. Please check credentials.")
        return session

    def projects(self):
        """Get projects authorized to user.

        List of projects user is authorized to see.

        Args:
            None

        Returns:
            An array of dict objects representing the projects the user has
            access to.

            For example:

        .. code-block:: python

            [
                {'active': True,
                'alias': 'Global',
                'code': 'PROJ-00000001',
                'id': 2,
                'isRestricted': False,
                'name': 'Global'}
            ]

        """
        url = "{}/api/projects".format(self.url)
        resp = self.session.get(url)
        resp.raise_for_status()
        return resp.json()

    def upload_files(self, files):
        """Upload a list of files to ACAS.

        Pass an array of files to ACAS and upload them to the server

        Args:
            files: An array of either string paths, Path objects (see
            :py:class:`pathlib.Path`), base64 encoded strings, or dicts
            with "name" and "data" (base64 encoded data) attributes.

        Returns:
            An object of responses from ACAS in the form:

        .. code-block:: python

            {'files': [
                {'name': '1_1_Generic.xlsx',
                'originalName': '1_1_Generic.xlsx',
                'size': 12887,
                'type': None,
                'deleteType': 'DELETE',
                'url': 'http://localhost:3000/dataFiles/1_1_Generic.xlsx',
                'deleteUrl': 'http://localhost:3000/uploads/1_1_Generic.xlsx'}
            ]}
        """
        filesToUpload = {}
        for file in files:
            if isinstance(file, Path):
                filesToUpload[str(file)] = file.open('rb')
            elif isBase64(file):
                filesToUpload[str("file")] = base64.decodebytes(file.encode())
            elif isinstance(file, dict):
                if isBase64(file['data']):
                    filesToUpload[file["name"]] = base64.decodebytes(
                        file["data"].encode())
                else:
                    filesToUpload[file["name"]] = file["data"]
            else:
                filesToUpload[str(file)] = file.open('rb')
        resp = self.session.post("{}/uploads".format(self.url),
                                 files=filesToUpload)
        # Close the open files
        for file in filesToUpload:
            # Check if the file is a file object
            if isinstance(filesToUpload[file], IOBase):
                filesToUpload[file].close()

        resp.raise_for_status()
        return resp.json()

    def add_files_to_lot(self, lot_corp_name, files):
        """Attach files to a lot.

        Attach files to a lot by passing a list of files and the lot's corp name.

        Args:
            lot_corp_name: The corp name of the lot to attach the files to.
            files: An array of dicts with the following keys
                "file" (Path): The file path to upload
                "type" (str): The type of file to upload
                "writeup" (str): The writeup for the file
        
        Returns: A dict with errors array and metalot dict object with the files attached to the lot

        """
        
        # Get the meta lot to make sure it exists
        meta_lot = self.get_meta_lot(lot_corp_name)

        # Upload the files to cmpd reg
        uploaded_files = self._upload_cmpdreg_files(lot_corp_name, files)

        # Attach the files to the lot
        for uploaded_file in uploaded_files:
            meta_lot['fileList'].append(uploaded_file)

        # Update the meta lot
        meta_lot_save_response = self.save_meta_lot(meta_lot)

        return meta_lot_save_response

    def add_file_to_lot(self, lot_corp_name, file, file_type, writeup=None):
        """Add a file to a lot

        Uploads and attaches a file to a lot

        Args:
            lot_corp_name (str): The corp name of the lot to attach the files to.
            file (Path): The file path to upload
            type (str): The type of file to upload
            writeup (str): The writeup for the file

        Returns: A dict with errors array and metalot dict object with the files attached to the lot

        """
        return self.add_files_to_lot(lot_corp_name, [{"file": file, "file_type": file_type, "writeup": writeup}])

    def _get_cmpdreg_files_content(self, cmpdreg_lot_file_list):
        """Get the content of cmpdreg files

        Pass the lot fileList array of array attribute of a cmpdreg metalot object to update or add the content attribute to each file

        Args:
            files: An array of dictionaries from the lot fileList attribute (attributes other than url are ignored)
                "url": The url of the file to download

        Returns:
            The the original array of dictionaries with the content attribute updated or added to each file
        """
        for file in cmpdreg_lot_file_list:
            resp = self.session.get("{}/cmpdreg/MultipleFilePicker/{}".format(self.url, file["url"]))
            resp.raise_for_status()
            file["content"] = resp.content
        return cmpdreg_lot_file_list

    def _upload_cmpdreg_files(self, lot_corp_name, files):
        """Upload a list of files to CmpdReg.

        Pass an array of files to ACAS and upload them to the creg server

        Args:
            lot_corp_name (str): The lot corp name to upload the file to
            files: An array of dicts with the following keys
                "file" (Path): The file path to upload
                "type" (str): The type of file to upload
                "writeup" (str): The writeup for the file

        Returns:
            An object of responses from ACAS in the form:

        .. code-block:: python

            [
                {
                    "description": "HPLC",
                    "ie": true,
                    "name": "dummy.pdf",
                    "size": 13264,
                    "subdir": "CMPD-0000002-001",
                    "type": "application/pdf",
                    "uploaded": true,
                    "url": "getFile?file
                }
            ]
        """
        filesToUpload = ()
        for file_dict in files:
            filesToUpload = filesToUpload + (('file[]', (file_dict["file"].name, file_dict["file"].open('rb'), 'application/octet-stream')),)
            filesToUpload = filesToUpload + (('description[]', (None, file_dict["file_type"])),)
            # If there is no writeup, or if we set the writeup to empty quotes "" the multipart form service in acas will fail
            # Setting a single space character sets the writeup to "" in the database
            if "writeup" not in file_dict or file_dict["writeup"] is None:
                writeup = " "
            else:
                writeup = file_dict["writeup"]
            filesToUpload = filesToUpload + (('writeup[]', (None, writeup)),)

        filesToUpload = filesToUpload + (('subdir', (None, lot_corp_name)),)
        filesToUpload = filesToUpload + (('ie', (None, 'true')),)

        headers = {
            'Accept': 'application/json'
        }
        resp = self.session.post("{}/cmpdreg/filesave".format(self.url),
                                 headers=headers,
                                 files=filesToUpload)

        resp.raise_for_status()
        return resp.json()

    def _upload_cmpd_reg_file(self, lot_corp_name, file, file_type, writeup=None):
        """Upload a file to CmpdReg.

        Upload a file to a specific lot corp name folder in cmpdreg
        Note: This does not attach the file to the specific lot it just uploads the file to that folder

        Args:
            lot_corp_name (str): The lot corp name to upload the file to
            file (Path): The file path to upload
            type (str): The type of file to upload
            writeup (str): The writeup for the file

        Returns:
            An object of responses from ACAS in the form:
        .. code-block:: python

            [
                {
                    "description": "HPLC",
                    "ie": true,
                    "name": "dummy.pdf",
                    "size": 13264,
                    "subdir": "CMPD-0000002-001",
                    "type": "application/pdf",
                    "uploaded": true,
                    "url": "getFile?fileUrl=%2Fhome%2Frunner%2Fbuild%2FprivateUploads%2Fcmpdreg%2FnoteBookFiles%2FnoteBook%2FCMPD-0000002-001%2Fdummy.pdf",
                    "writeup": "Nothing"
                }
            ]
        """
        response = self._upload_cmpdreg_files(lot_corp_name, [{"file": file, "file_type": file_type, "writeup": writeup}])
        return response[0]

    def get_meta_lot(self, lot_corp_name):
        """Get metalot by lot corp name
         Granted read permission on a lot if one of these is true:
            1. The user is the owner of the lot (chemist or recorded by)
            2. The user has access to the project the lot is associated with
            3. The user is a cmpdreg admin

        Args:
            lot_corp_name (str): A lot corp name

        Returns: Returns a dict meta lot object
        """
        resp = self.session.get("{}/cmpdreg/metalots/corpName/{}/"
                                .format(self.url, lot_corp_name))
        if resp.status_code == 500 or resp.status_code == 404:
            return None
        resp.raise_for_status()
        return resp.json()

    def save_meta_lot(self, meta_lot):
        """Save a meta lot to the server
         If updating a saved lot permissions are granted if one of these is true:
            1. Edit my lots is configured to true on the system and..
                a. The user is the owner of the lot (chemist or recorded by)
                b. The user has access to the project the lot is associated with
            2. The user is a cmpdreg admin

        Args:
            meta_lot (dict): A meta lot

        Returns: Returns a dict meta lot object
        """
        resp = self.session.post("{}/cmpdreg/metalots".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(meta_lot))
        resp.raise_for_status()
        return resp.json()

    def cmpd_search(self, corpNameList="", corpNameFrom="", corpNameTo="",
                    aliasContSelect="contains", alias="", dateFrom="",
                    dateTo="", searchType="substructure", percentSimilarity=90,
                    chemist="anyone", maxResults=100, molStructure="", projectCodes=None
                    ):
        if isinstance(corpNameList, list):
            corpNameList = ",".join(corpNameList)
        # Put all local variables into a dictionary
        search_request = dict(locals())
        del search_request['self']

        if searchType not in VALID_STRUCTURE_SEARCH_TYPES:
            raise ValueError("cmpd_search: searchType must be one of %r."
                             % VALID_STRUCTURE_SEARCH_TYPES)
        return self.cmpd_search_request(search_request)

    def cmpd_structure_search(self, searchType="substructure", percentSimilarity=90,
                    maxResults=100, molStructure=""
                    ):
        search_request = dict(locals())
        del search_request['self']

        if searchType not in VALID_STRUCTURE_SEARCH_TYPES:
            raise ValueError("cmpd_search: searchType must be one of %r."
                             % VALID_STRUCTURE_SEARCH_TYPES)
        return self.cmpd_structure_search_request(search_request)

    def get_all_lots(self, project_codes=None):
        """Get all lots

        Get all lots the currently logged in user is allowed to access

        Args:
            project_codes (list): A list of project codes to filter by

        Returns: Returns an array of dict objects
            id (id): the lot corp name
            lotCorpName (str): the lot corp name
            lotNumber (int): the lot number
            parentCorpName (str): the lots parent corp name
            registrationDate (int): the registration date
            project (str): the lots project
        """
        resp = self.session.get("{}/cmpdReg/parentLot/getAllAuthorizedLots"
                                .format(self.url))
        resp.raise_for_status()

        lots = resp.json()

        if project_codes is not None:
            lots = [lot for lot in lots if lot["project"] in project_codes]
        return lots


    def get_all_meta_lots(self, project_codes=None):
        """Get all meta lots

        Get all meta lots the currently logged in user is allowed to access

        Args:
            project_codes (list): A list of project codes to filter by

        Returns: Returns an array of dict objects respresenting meta lots
        """
        
        # Get all lots
        lots = self.get_all_lots(project_codes=project_codes)

        # Get meta lots
        meta_lots = self.get_meta_lots_by_lot_corp_names([lot["lotCorpName"] for lot in lots])

        return meta_lots
        
    def get_meta_lots_by_lot_corp_names(self, lot_corp_names, max_workers=10):
        """Get meta lots by lot corp names

        Get meta lots by lot corp names

        Args:
            lot_corp_names (list): A list of lot corp names

        Returns: Returns an array of dict objects respresenting meta lots
        """
        meta_lots = []

        # Call get_meta_lot for each lot corp name
        # Run max_workers requests in parallel
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for meta_lot in executor.map(self.get_meta_lot, lot_corp_names):
                meta_lots.append(meta_lot)
        return meta_lots

    def cmpd_search_request(self, search_request):
        search_request["loggedInUser"] = self.username
        if("corpNameList" not in search_request):
            search_request["corpNameList"] = ""
        if("corpNameFrom" not in search_request):
            search_request["corpNameFrom"] = ""
        if("corpNameTo" not in search_request):
            search_request["corpNameTo"] = ""
        if("aliasContSelect" not in search_request):
            search_request["aliasContSelect"] = "contains"
        if("alias" not in search_request):
            search_request["alias"] = ""
        if("dateFrom" not in search_request):
            search_request["dateFrom"] = ""
        if("dateTo" not in search_request):
            search_request["dateTo"] = ""
        if("searchType" not in search_request):
            search_request["searchType"] = "substructure"
        if("dateFrom" not in search_request):
            search_request["dateFrom"] = ""
        if("percentSimilarity" not in search_request):
            search_request["percentSimilarity"] = 90
        if("percentSimilarity" not in search_request):
            search_request["percentSimilarity"] = 90
        if("chemist" not in search_request):
            search_request["chemist"] = "anyone"
        if("maxResults" not in search_request):
            search_request["maxResults"] = 100
        if("projectCodes" in search_request and search_request["projectCodes"] is None):
            del search_request["projectCodes"]

        resp = self.session.post("{}/cmpdreg/search/cmpds".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(search_request))
        resp.raise_for_status()
        return resp.json()

    def cmpd_structure_search_request(self, search_request):
        if("searchType" not in search_request):
            search_request["searchType"] = "substructure"
        if("percentSimilarity" not in search_request):
            search_request["percentSimilarity"] = 90

        resp = self.session.post("{}/cmpdreg/structuresearch/".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(search_request))
        resp.raise_for_status()
        return resp.json()

    def export_cmpd_search_results(self, search_results):
        """Export an sdf of compound search results.

        Given a search results dict object this will export a list of matching
        compounds to an SDF file.

        Args:
            search_results: Dict object

        Search criteria for lots

        .. code-block:: python

            search_results = {
                "foundCompounds": [
                    {
                        "lotIDs": [
                            {
                                "corpName": "CMPD-0000001-001"
                            }
                        ],
                    }
                ]
            }

        Full List of potential search results

        .. code-block:: python

            search_results = {
                "foundCompounds": [
                    {
                        "corpName": "CMPD-0000001",
                        "corpNameType": "Parent",
                        "lotIDs": [
                            {
                                "corpName": "CMPD-0000001-001",
                                "lotNumber": 1,
                                "registrationDate": "01/29/2020",
                                "synthesisDate": "01/29/2020"
                            }
                        ],
                        "molStructure": "MOLFILE STRUCTURE"
                        "parentAliases": [],
                        "stereoCategoryName": "Achiral",
                        "stereoComment": ""
                    }
                ],
                "lotsWithheld": False
            }

        Returns:
            An object of responses from ACAS in the form:

        .. code-block:: python

            {'reportFilePath':
                '/dataFiles/exportedSearchResults/2020_02_12_1581466283857_searchResults.sdf',
            'summary': 'Successfully exported 1 lots.'
            }

        See the output of :func:`get_sdf_file_for_lots` for SDF details.
        """
        resp = self.session.post("{}/cmpdReg/export/searchResults".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(search_results))
        resp.raise_for_status()
        return resp.json()

    def get_sdf_file_for_lots(self, lots):
        """Get an SDF file object from an array of lot corp names

        Given an array of lots this function fetches and SDF file for those
        lots with their lot and parent attibutes

        Args:
            lots: Array of lot corp names
        Returns:
            ACAS file dict object with sdf content

        Output object structure:

        .. code-block:: python

            {'content-type': 'application/octet-stream',
            'content-length': '1642',
            'last-modified': 'Wed, 12 Feb 2020 00:49:09 GMT',
            'name': '2020_02_12_1581468549556_searchResults.sdf',
            'content': b'SDFILE CONTENT'}

        SDF Attributes:


        * Amount Units Code (str): amount units
        * Buid (int): legacy field not used
        * Bulk Load File (str): Bulk load file that the lot came from
        * Chemist (str): the lot chemist
        * Lot Corp Name (str):
        * Is Virtual (bool): Is this a virtual compound
        * Lot Mol Weight (float): Lot weight (includes salt weight)
        * Lot Number (int):
        * Project (str):
        * Registration Date (str): Lot registration date in the
          format "2020-02-11"
        * Lot Registered By (str):
        * Salt Form Corp Name (str): Salt form name
        * Parent Corp Name (str):
        * Parent Number (int): Internal parent identifier
        * Parent Stereo Category (str):
        * Parent Mol Weight (float):
        * Parent Exact Mass (float):
        * Parent Mol Formula (str):
        * Parent Registration Date (str): Parent registration date in the
          format "2020-02-11"
        * Parent Registered By (str):

        """
        lotIds = []
        for lot in lots:
            lotIds.append({"corpName": lot})
        search_results = {
            "foundCompounds": [
                {
                    "lotIDs": lotIds,
                }
            ]
        }
        search_results_export = self.export_cmpd_search_results(search_results)
        search_results_file = self.get_file(search_results_export[
                                            'reportFilePath'])
        return search_results_file

    def write_sdf_file_for_lots(self, lots, dir_or_file_path):
        """Get and write an SDF from lot corp name array

        Given an array of lots this function fetches and SDF file for those
        lots with their lot and parent attibutes and write an SDF file

        Args:
            lots: Array of lot corp names
        Returns:
            ACAS file dict object with sdf content

        See the output of :func:`get_sdf_file_for_lots` for SDF details.
        """
        sdf_file = self.get_sdf_file_for_lots(lots)
        file_path = self.write_file(sdf_file, dir_or_file_path)
        return file_path

    def get_file(self, file_path, parse_content=True):
        """Get a file from ACAS

        Get a file from ACAS

        Note:
            If behind a proxy some fields (denoted with a ``*``) may not be
            filled

        Args:
            file_path (str): A path to a file known to exist in ACAS

        Returns: ACAS file dict object with file content
            content-type* (str): Content type of the file
            content-length* (int): Content length in bytes of the file
            last-modified* (str): Date file was last mofied
            name (str): Name of the file
            content (str): Content of the file
            parsed_content (depends): Parsed content of the file
        """
        resp = self.session.get("{}{}".format(self.url, file_path))
        resp.raise_for_status()
        name = PurePath(Path(file_path)).name
        return_dict = {
            'content-type': resp.headers.get('content-type', None),
            'content-length': resp.headers.get('content-length', None),
            'last-modified': resp.headers.get('Last-modified', None),
            'name': name,
            'content': resp.content}

        # Get file extension
        file_extension = PurePath(Path(file_path)).suffix
        
        if parse_content:
            try:
                # This function returns None if the file extension is not
                # recognized
                return_dict['parsed_content'] = parse_file(
                    resp.content, file_extension)
            except Exception as e:
                # In case there is an error parsing the file, just return the
                # None
                logger.warning("Could not parse file: {}".format(e))
                return_dict['parsed_content'] = None

        return return_dict

    def protocol_search(self, search_term):
        """Search for protocols by search term

        Get an array of protocols given a protocol search term string

        Args:
            searchTerm (str): A protocol search term

        Returns: Returns an array of protocols
        """
        resp = self.session.get("{}/api/protocols/genericSearch/{}"
                                .format(self.url, search_term))
        resp.raise_for_status()
        return resp.json()

    def get_protocols_by_label(self, label):
        """Get all experiments for a protocol from a protocol label

        Get an array of experiments given a protocol label

        Args:
            label (str): A protocol label

        Returns: Returns an array of experiments
        """
        resp = self.session.get("{}/api/getProtocolByLabel/{}"
                                .format(self.url, label))
        resp.raise_for_status()
        return resp.json()

    def get_experiments_by_protocol_code(self, protocol_code):
        """Get all experiments for a protocol from a protocol code

        Get an array of experiments given a protocol code

        Args:
            protocol_code (str): A protocol code

        Returns: Returns an array of experiments
        """
        resp = self.session.get("{}/api/experiments/protocolCodename/{}".
                                format(self.url, protocol_code))
        if resp.status_code == 500:
            return None
        else:
            resp.raise_for_status()
        return resp.json()

    def get_experiment_by_name(self, experiment_name):
        """Get an experiment from experiment name

        Get an experiment given an experiment name

        Args:
            experiment_name (str): An experiment name

        Returns: Returns an experiment object or None if the experiment not found
        """

        resp = self.session.get("{}/api/experiments/experimentName/{}".
                                format(self.url, experiment_name))
        if resp.status_code == 500:
            return None
        resp.raise_for_status()
        return resp.json()

    def get_experiment_by_code(self, experiment_code, full = False):
        """Get an experiment from an experiment code

        Get an experiment given an experiment code

        Args:
            experiment_code (str): An experiment code code
            full (bool): If true, return the full experiment object

        Returns: Returns an experiment object
        """

        params = {}
        if full:
            params = {**params, 'fullObject': True}
        resp = self.session.get("{}/api/experiments/codename/{}".
                                format(self.url, experiment_code),
                                params = params)
        if resp.status_code == 500:
            return None
        else:
            resp.raise_for_status()
        return resp.json()

    def get_source_file_for_experient_code(self, experiment_code):
        """Get the source file for an experiment

        Gets the experiment loader file for ACAS

        Note:
            If behind a proxy some fields (denoted with a ``*``) may not be
            filled

        Args:
            file_path (str): A path to a file known to exist in ACAS

        Returns: ACAS file dict object with file content
            content-type* (str): Content type of the file
            content-length* (int): Content length in bytes of the file
            last-modified* (str): Date file was last mofied
            name (str): Name of the file
            content (str): Content of the file

        """
        experiment = self.get_experiment_by_code(experiment_code)
        if not experiment:
            return None
        source_file = get_entity_value_by_state_type_kind_value_type_kind(
            experiment,
            "metadata",
            "experiment metadata",
            "fileValue",
            "source file")
        file = None
        if source_file and "fileValue" in source_file:
            file = self.get_file("/dataFiles/{}"
                                 .format(source_file['fileValue']))
        return file

    def write_file(self, file, dir_or_file_path):
        dir_or_file_path = Path(dir_or_file_path)
        if dir_or_file_path.is_dir():
            file_path = dir_or_file_path.joinpath(file['name'])
        else:
            file_path = dir_or_file_path
        mode = "w"
        if type(file['content']) is bytes:
            mode = "wb"
        with open(file_path, mode) as f:
            f.write(file['content'])
        return file_path

    def write_source_file_for_experient_code(self, experiment_code, dir):
        source_file = self.get_source_file_for_experient_code(experiment_code)
        file_path = None
        if source_file:
            file_path = self.write_file(source_file, dir)
        return file_path

    def register_sdf_request(self, data):
        resp = self.session.post("{}/api/cmpdRegBulkLoader/registerCmpds"
                                 .format(self.url),
                                 headers={'Content-Type': 'application/json'},
                                 data=json.dumps(data))
        resp.raise_for_status()
        return resp.json()
    
    def _validate_sdf_request(self, data):
        resp = self.session.post("{}/api/cmpdRegBulkLoader/validateCmpds"
                                 .format(self.url),
                                 headers={'Content-Type': 'application/json'},
                                 data=json.dumps(data))
        resp.raise_for_status()
        return resp.json()

    def register_sdf(self, file, userName, mappings, prefix=None, dry_run=False):
        files = self.upload_files([file])
        if "files" not in files and len(files["files"]) != 1:
            raise RuntimeError("Failed to upload file when trying to register SDF.")
        else:
            uploaded_file = files['files'][0]

        request = {
            "fileName": uploaded_file["name"],
            "userName": userName,
            "mappings": mappings,
        }

        # If original name is returned from the fle upload service, pass it as part of the request so it is preserved by the server
        if "originalName" in uploaded_file:
            request["originalFileName"] = uploaded_file["originalName"]

        if prefix:
            request["labelPrefix"] = {
                "name": prefix,
                "labelPrefix": prefix,
                "labelTypeAndKind": "id_corpName",
                "thingTypeAndKind": "parent_compound"
            }
        if dry_run:
            response = self._validate_sdf_request(request)
        else:
            response = self.register_sdf_request(request)
        report_files = []
        for file in response[0]['reportFiles']:
            filePath = "/dataFiles/cmpdreg_bulkload/{}".format(
                PurePath(Path(file)).name)
            report_files.append(self.get_file(filePath))
        return {"id": response[0]['id'],
                "summary": response[0]['summary'],
                "results": response[0]['results'],
                "report_files": report_files}

    def experiment_loader_request(self, data):
        resp = self.session.post("{}/api/genericDataParser".format(self.url),
                                 headers={'Content-Type': 'application/json'},
                                 data=json.dumps(data))
        resp.raise_for_status()
        return resp.json()

    def _dose_response_fit_request(self, dose_response_request_dict):
        """ Send a dose response fit request to ACAS
        
        This is a private method that is used to send a dose response fit json request to ACAS. It is not intended to be used directly.

        Args:
            dose_response_request_dict (dict): A dictionary containing the request parameters for the dose response fit request.
        """
    
        resp = self.session.post("{}/api/doseResponseCurveFit".format(self.url),
                                 headers={'Content-Type': 'application/json'},
                                 data=json.dumps(dose_response_request_dict))
        resp.raise_for_status()
        return resp.json()

    def experiment_loader(self, data_file, user, dry_run, report_file="",
                          images_file="", validate_dose_response_curves=True):
        """Load an experiment
        
        Load an experiment into ACAS.

        Args:
            data_file (str): A path to an experiment loader formatted file
            user (str): A username
            dry_run (bool): If true, then validate but do not load the data into the database
            report_file (str): A path to a report file (optional)
            images_file (str): A path to an images file (optional)
        """
        data_file_path = self.upload_files([data_file])['files'][0]["name"]
        report_file_path = ""
        if report_file and report_file != "":
            report_file_path = self.upload_files([report_file])['files'][0]["name"]
        images_file_path = ""
        if images_file and images_file != "":
            images_file_path = self.upload_files([images_file])['files'][0]["name"]
        request = {"user": user,
                   "fileToParse": data_file_path,
                   "reportFile": report_file_path,
                   "imagesFile": images_file_path,
                   "moduleName": None if validate_dose_response_curves else "DoseResponseDataParserController",
                   "dryRunMode": dry_run}
        resp = self.experiment_loader_request(request)
        return resp

    def dose_response_experiment_loader(self, model_fit_type, fit_settings, **kwargs):
        """Dose response experiment loader
        
        Args:
            model_fit_type (str): The type of model fit to perform
            fit_settings (dict): The settings for the model fit
            **kwargs: All required arguments to pass to the experiment loader (e.g. data_file, user, dry_run = True/False)

        Returns:
            dict: The response from the experiment loader and doseresponse fit request
             
            Example:

                {
                    "experiment_loader_response": experiment_loader_response_resp_dict,
                    "dose_response_fit_response": dose_response_fit_response_resp_dict
                }
    
        Example:
            request = {
                "data_file": data_file_to_upload,
                "user": "bob",
                "dry_run": True,
                "model_fit_type": "4 parameter D-R",
                "fit_settings": {
                    "smartMode":True,
                    "inactiveThresholdMode":True,
                    "inactiveThreshold":20,
                    "theoreticalMaxMode":False,
                    "theoreticalMax":None,
                    "inverseAgonistMode":False,
                    "max":{
                        "limitType":"none"
                    },
                    "min":
                        {
                            "limitType":"none"
                    },
                    "slope":{
                        "limitType":"none"
                    },
                    "baseline":{
                        "value":0
                    }
                }
            }
            response = client.\
                dose_response_experiment_loader(**request)
        """

        resp = self.experiment_loader(validate_dose_response_curves=False, **kwargs)
        response = {
            "experiment_loader_response": resp,
            "dose_response_fit_response": None
        }
        if resp['hasError'] == False and resp['commit'] == True:
            request = {
                "experimentCode": resp['results']['experimentCode'],
                "modelFitType": model_fit_type,
                "testMode": False,
                "user": kwargs['user'],
                "inputParameters": json.dumps(fit_settings)
            }
            dose_response_resp = self._dose_response_fit_request(request)
            response["dose_response_fit_response"] = dose_response_resp
        return response

    def delete_experiment(self, idOrCode):
        """Delete an experiment

        Deletes an experiment.  If a code name is given, the experiment is
        first retrieved by code name and then deleted by id.

        Args:
            idOrCode (int or str): An experiment id or an experiment code name

        Returns: Dict object with the experiment status value of the deleted
                 experiment

        """
        if isinstance(idOrCode, str):
            experiment = self.get_experiment_by_code(idOrCode)
            if experiment is None:
                return None
            else:
                idOrCode = experiment['id']

        resp = self.session.delete("{}/api/experiments/{}".
                                   format(self.url, idOrCode)
                                   )
        resp.raise_for_status()
        return resp.json()

    def experiment_search(self, query, project_codes=None):
        """Search for experiments by search term

        Get an array of experiments given an experiment search term string and optional project code(s) filter

        Args:
            query (str): An experiment search term
            project_codes (str list): A list of project codes to filter by

en array of protocols
        """

        params = {}
        if project_codes is not None:
            # Convert to a comma separated string and url encode
            params['projectCodes'] = ','.join(project_codes)

        resp = self.session.get("{}/api/experiments/genericSearch/{}/"
                                .format(self.url, query),
                                params=params)

        resp.raise_for_status()
        return resp.json()

    def get_cmpdreg_bulk_load_files(self):
        """Get cmpdreg bulk load files

        Gets a list of all cmpdreg bulk files on the system

        Returns: An array of dict objects
            fileDate (int): The epoch date the file was registered
            fileName (str): The name of the file
            fileSize (int): Size in bytes of the file
            id (int): The file id
            numberOfMols (int): Number of mols registered by this file
            recordedBy (str): Username of the user who registered the file
            version (int): The file version number

        """
        resp = self.session.get("{}/api/cmpdRegBulkLoader/getFilesToPurge"
                                .format(self.url))
        return resp.json()

    def check_cmpdreg_bulk_load_file_dependency(self, id):
        """Check cmpdreg bulk load file dependencies

        Check for dependencies of cmpdreg bulk load file

        Args:
            id (int): A bulk load file id

        Returns: Dict object with file content
            canPurge (bool): Can this file be purged
            summary (str): An html formatted summary of the dependencies

        """
        request = {
            "fileInfo": {
                "id": id
            }
        }
        resp = self.session.post("{}/api/cmpdRegBulkLoader/checkFileDependencies".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(request))
        if resp.text == '"Error"':
            return None
        return resp.json()

    def get_lot_corp_names_by_bulk_load_file(self, id):
        """Get an array of lot corp names from a bulk load file

        Args:
            id (int): A bulk load file id

        Returns: An array of lot corp names

        """
        resp = self.session.get("{}/api/cmpdRegBulkLoader/getLotsByBulkLoadFileID/{}".
                                format(self.url,
                                       id))
        resp.raise_for_status()
        if resp.text == '"Error"':
            return None
        return resp.json()

    def get_sdf_by_bulk_load_file(self, id):
        """Get an SDF file from a bulk load file id

        Args:
            id (int): A bulk load file id

        Returns: (str) representaiton of the SDF file

        """
        resp = self.session.get("{}/api/cmpdRegBulkLoader/getSDFFromBulkLoadFileId/{}".
                                format(self.url,
                                       id))
        resp.raise_for_status()
        if resp.text == '"Error"':
            return None
        return resp.text

    def purge_cmpdreg_bulk_load_file(self, id):
        """Purge a cmpdreg bulk load file

        Purges a cmpdreg bulk load file

        Args:
            id (int): A bulk load file id

        Returns: Dict object with file content
            fileName (str): The name of the file that was purged
            success (bool): Did the file purge successfully
            summary (str): An html formatted summary of the purge results

        """
        request = {
            "fileInfo": {
                "id": id
            }
        }
        resp = self.session.post("{}/api/cmpdRegBulkLoader/purgeFile".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(request))
        if resp.text == '"Error"':
            return None
        return resp.json()

    def get_ls_thing(self, ls_type, ls_kind, code_name, nestedfull=True):
        """
        Get a models.LsThing object by ls_type, ls_kind, and code_name

        Args:
            ls_type (str): Type of ls thing
            ls_kind (str): Kind of ls thing
            code_name (str): Code name of ls thing

        """
        resp = self.session.get("{}/api/things/{}/{}/{}".
                                format(self.url,
                                       ls_type,
                                       ls_kind,
                                       code_name),
                                params={'nestedfull': nestedfull})
        if resp.status_code == 500:
            return None
        else:
            resp.raise_for_status()
        return resp.json()
    
    def delete_ls_thing(self, ls_type, ls_kind, code_name, format):
        """
            Deletes a models.LsThing object by ls_type, ls_kind, and code_name
            Args:
                ls_type (str): Type of ls thing
                ls_kind (str): Kind of ls thing
                code_name (str): Code name of ls thing
                format (str)
        """
        if not format:
            format = 'nestedfull'
        resp = self.session.delete(
            "{}/api/things/{}/{}/{}".format(self.url, ls_type, ls_kind,
                                            code_name),
            params={format: True})
        if resp.status_code == 500:
            return None
        else: 
            resp.raise_for_status()
        return resp

    def save_ls_thing(self, ls_thing):
        """
        Persist a models.LsThing object to ACAS

        Args:
            ls_thing (dict): A dict object representing an ls thing

        Returns: Dict object representing a saved ls_thing
        """

        resp = self.session.post("{}/api/things/{}/{}".
                                 format(self.url,
                                        ls_thing["lsType"],
                                        ls_thing["lsKind"]),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(ls_thing))
        resp.raise_for_status()
        return resp.json()

    def get_ls_things_by_codes(self, ls_type, ls_kind, code_name_list,
                               nestedfull=True):
        """
        Get a list of ls thing dict objects from a list of their code_names

        Args:
            ls_type (str): ls_type for all things to retrieve
            ls_kind (str): ls_kind for all things to retrieve
            code_name_list (str list): list of str code_names
        """
        params = {}
        if nestedfull:
            params = {**params, 'with': 'nestedfull'}
        resp = self.session.post("{}/api/things/{}/{}/codeNames/jsonArray".
                                 format(self.url,
                                        ls_type,
                                        ls_kind),
                                 params=params,
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(code_name_list))
        if resp.status_code == 500:
            return None
        else:
            resp.raise_for_status()
        return resp.json()

    def get_ls_things_by_type_and_kind(self, ls_type, ls_kind,
                                       format='stub'):
        """
        Get a list of ls thing dict objects from ls_type and ls_kind

        Args:
            ls_type (str): ls_type for all things to retrieve
            ls_kind (str): ls_kind for all things to retrieve
        """
        allowedFormats = {'codetable', 'stub'}
        if format not in allowedFormats:
            raise ValueError("format must be one of %s." % allowedFormats)
        params = {format: '1'}
        resp = self.session.get("{}/api/things/{}/{}".
                                format(self.url,
                                       ls_type,
                                       ls_kind),
                                params=params,
                                headers={'Content-Type': "application/json"})
        if resp.status_code == 500:
            return None
        else:
            resp.raise_for_status()
        return resp.json()

    def save_ls_thing_list(self, ls_thing_list):
        """
        Save a list of ls thing dict objects

        Args:
            ls_thing_list (str): list of ls_thing dict objects
        """
        resp = self.session.post("{}/api/bulkPostThingsSaveFile".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(ls_thing_list))
        resp.raise_for_status()
        return resp.json()

    def update_ls_thing_list(self, ls_thing_list):
        """
        Update a list of ls thing dict objects

        Args:
            ls_thing_list (str): list of ls_thing dict objects
        """
        # TODO: generate a transaction
        resp = self.session.put("{}/api/bulkPutThingsSaveFile".
                                format(self.url),
                                headers={'Content-Type': "application/json"},
                                data=json.dumps(ls_thing_list))
        resp.raise_for_status()
        return resp.json()

    def get_thing_codes_by_labels(self, thing_type, thing_kind, labels_or_codes, label_type=None, label_kind=None):
        """
        Get a list of thing codes by providing a list of labels

        Args:
            labels_or_codes (str list): list of str labels or codes
            thing_type (str): ls_type for all things to retrieve
            thing_kind (str): ls_kind for all things to retrieve
            label_type (str): label_type to limit label searches
            label_kind (str): label_kind to limit label searches
        Returns:
            ref_name_lookup_results: list of objects with
                requestName (str): input label string
                preferredName (str): LsThing preferred label string
                referenceName (str): LsThing code name string
        """
        request = {
            'thingType': thing_type,
            'thingKind': thing_kind,
            'labelType': label_type,
            'labelKind': label_kind,
            'requests': [
                {"requestName": request} for request in labels_or_codes
            ]
        }

        resp = self.session.post("{}/api/getThingCodeByLabel".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(request))
        if resp.status_code == 500:
            return None
        else:
            resp.raise_for_status()
        resp_object = resp.json()
        if resp_object == 'error trying to lookup lsThing name':
            raise RuntimeError("Failed to get things, please see acas logs.")
        results = resp_object['results']
        return results

    def get_saved_entity_codes(self, ls_type, ls_kind, id_list, label_type=None, label_kind=None):
        """
        Query ACAS to determine which identifiers (labels) are already saved

        Args:
            ls_type (str): LsThing lsType to query for
            ls_kind (str): LsThing lsKind to query for
            id_list (str): list of identifier strings
            label_type (str): label_type to limit label searches
            label_kind (str): label_kind to limit label searches
        Returns:
            saved_codes (dict): dict of identifier : LsThing codeName for previously saved entities
            missing_ids (list): list of identifiers that were not found to be previously saved
        """
        # Query ACAS for list of identifiers
        ref_name_lookup_results = self.get_thing_codes_by_labels(
            ls_type, ls_kind, id_list, label_type, label_kind)
        # Parse results into found and not found
        saved_codes = {}
        missing_ids = []
        for res in ref_name_lookup_results:
            ident = res['requestName']
            if res['referenceName'] and len(res['referenceName']) > 0:
                saved_codes[ident] = res['referenceName']
            else:
                missing_ids.append(ident)
        return saved_codes, missing_ids

    def advanced_search_ls_things(self, ls_type, ls_kind, search_string,
                                  value_listings=[], label_listings=[],
                                  first_itx_listings=[], second_itx_listings=[],
                                  codes_only=False,
                                  max_results=1000, combine_terms_with_and=False,
                                  format='stub', return_listings=None):
        """
        Query ACAS for deeply specified conditions

        Args:
            ls_type (str): LsThing lsType to match
            ls_kind (str): LsThing lsKind to match
            search_string (str): str to match on or compare to
            value_listings (list): list of dicts of a structure like:
                {
                    "stateType": "metadata",
                    "stateKind": "pdb",
                    "valueType": "stringValue",
                    "valueKind": "librarian search status",
                    "operator": "="
                }
            combine_terms_with_and (bool): Whether to combine terms with 'and'
            format (str): ACAS format to fetch data in
            return_listings (dict): Used when format = "flat". Return only the defined values or label listings and thing attributes in key value pair format. 
                {
                    "thingValues": [
                        {
                            # Code value return listing
                            "key": return_status_key,
                            "stateType": "metadata",
                            "stateKind": "project metadata",
                            "valueType": "codeValue",
                            "valueKind": PROJECT_STATUS
                        },
                        {
                            # String value return listing
                            "key": return_description_key,
                            "stateType": "metadata",
                            "stateKind": "project metadata",
                            "valueType": "stringValue",
                            "valueKind": DESCRIPTION_KEY
                        },
                        {
                            # Date value return listing
                            "key": return_startdate_key,
                            "stateType": "metadata",
                            "stateKind": "project metadata",
                            "valueType": "dateValue",
                            "valueKind": START_DATE
                        },
                        {
                            # Label return listing
                            "key": return_name_key,
                            "labelType": "name",
                            "labelKind": PROJECT_NAME,
                        },
                        {
                            # Numeric Value return listing
                            "key": return_project_number_key,
                            "stateType": "metadata",
                            "stateKind": "project metadata",
                            "valueType": "numericValue",
                            "valueKind": PROJECT_NUMBER
                        }
                    ],
                    "thingAttributes": ["codeName", "id", "modifiedDate"]
                }
        Returns:
            if codes_only:
                list of code_name strings
            otherwise:
                list of LsThing objects
        """
        request = {
            'queryString': search_string,
            'queryDTO': {
                'maxResults': max_results,
                'lsType': ls_type,
                'lsKind': ls_kind,
                'values': value_listings,
                'labels': label_listings,
                'firstInteractions': first_itx_listings,
                'secondInteractions': second_itx_listings,
                'combineTermsWithAnd': combine_terms_with_and,
            },
            'returnDTO': return_listings

        }
        params = {}
        if codes_only:
            format = 'codetable'
        params['format'] = format
        resp = self.session.post('{}/api/advancedSearch/things/{}/{}'.
                                 format(self.url,
                                        ls_type,
                                        ls_kind),
                                 data=json.dumps(request),
                                 headers={'Content-Type': "application/json"},
                                 params=params)
        result = resp.json()
        if type(result) is not dict:
            msg = 'Caught error response from {}: {}'.format(
                '/api/advancedSearch/things', result)
            logger.error(msg)
            raise ValueError(msg)
        results = result['results']
        if codes_only:
            return [res['code'] for res in results]
        else:
            return results

    def create_label_sequence(self, labelPrefix, startingNumber, digits,
                              labelSeparator, labelTypeAndKind=None, thingTypeAndKind=None,
                              labelSequenceRoles=[]):
        """
        Create a label sequence

        Args:
            labelPrefix (str): Prefix of the label
            startingNumber (str): Set to 0 for the first number to be 1
            digits (str): The number of leading zeros to add to the label sequence when formatting (e.g. CMPD-0000007 would be digts: 7 )
            labelTypeAndKind (str): the label type and kind associated with this sequence (used for finding all labels of a specific label type and kind in some interfaces)
            thingTypeAndKind (str): the thing type and kind associated with this sequence (used for finding all labels of a specific thing type and kind in some interfaces)
            labelSequenceRoles (list): the registered role to associate with this label (used for limiting access to specific label sequences in some interfaces)

        Returns:
            a dict object representing the new label sequence
        """
        request = {
            'labelPrefix': labelPrefix,
            'startingNumber': startingNumber,
            'digits': digits,
            'labelSeparator': labelSeparator,
            'labelTypeAndKind': labelTypeAndKind,
            'thingTypeAndKind': thingTypeAndKind,
            'labelSequenceRoles': labelSequenceRoles
        }
        resp = self.session.post("{}/api/labelsequences/".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(request))
        resp.raise_for_status()
        return resp.json()

    def get_all_label_sequences(self):
        """
        Get all label sequences (limited to those authorized by logged in user roles)

        Returns:
            a list of dict objects representing the labelSequence
        """

        json = self.get_label_sequence_by_types_and_kinds()
        return json

    def get_label_sequence_by_types_and_kinds(self, labelTypeAndKind=None, thingTypeAndKind=None):
        """
        Get label sequence by types and kinds (limited to those authorized by logged in user roles)

        Args:
            labelPrefix (str): Prefix of the label
            startingNumber (str): Set to 0 for the first number to be 1
            digits (str): The number of leading zeros to add to the label sequence when formatting (e.g. CMPD-0000007 would be digts: 7 )
            labelTypeAndKind (str): the label type and kind associated with this sequence (used for finding all labels of a specific label type and kind in some interfaces)
            thingTypeAndKind (str): the thing type and kind associated with this sequence (used for finding all labels of a specific thing type and kind in some interfaces)
            labelSequenceRoles (list): the registered role to associate with this label (used for limiting access to specific label sequences in some interfaces)

        Returns:
            a list of dict objects representing the labelSequence
        """
        params = {}
        if labelTypeAndKind:
            params = {**params, 'labelTypeAndKind': labelTypeAndKind}
        if thingTypeAndKind:
            params = {**params, 'thingTypeAndKind': thingTypeAndKind}
        resp = self.session.get("{}/api/labelSequences/getAuthorizedLabelSequences".
                                format(self.url),
                                params=params)
        resp.raise_for_status()
        return resp.json()

    def get_labels(self, labelTypeAndKind, thingTypeAndKind, numberOfLabels):
        """
        Get next n labels from label sequence prefix

        Args:
            labelTypeAndKind (str): Prefix of the registered label (see create_label_sequence)
            numberOfLabels (int): Number of labels to fetch

        Returns:
            a list of dict objects representing the labelSequence
        """
        request = {
            'labelTypeAndKind': labelTypeAndKind,
            'thingTypeAndKind': thingTypeAndKind,
            'numberOfLabels': numberOfLabels
        }
        resp = self.session.post("{}/api/getNextLabelSequence".
                                 format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(request))
        resp.raise_for_status()
        return resp.json()

    def get_all_ddict_values(self):
        """
        Get all ddict values

        Returns:
            a list of dict objects representing the ddict value (aka code value)
        """
        all_values = self.get_ddict_values_by_type_and_kind()
        return all_values

    def get_ddict_values_by_type_and_kind(self, codeType=None, codeKind=None):
        """
        Get ddict values

        Returns:
            a list of dict objects representing the ddict value (aka code value)
        """
        path = "/api/codetables"
        if codeType and codeKind:
            path = "{}/{}/{}".format(path, codeType, codeKind)
        resp = self.session.get("{}{}".
                                format(self.url, path))
        resp.raise_for_status()
        return resp.json()

    def get_blob_data_by_value_id(self, valueId):
        """
        Get blob data by value id
        Args:
            valueId (int): A known value id to fetch from the database that is stored as a blobValue lsType

        Returns:
            (bytes): representing the blob value
        """

        resp = self.session.get("{}/api/thingvalues/downloadThingBlobValueByID/{}"
                                .format(self.url, valueId))
        resp.raise_for_status()
        return resp.content
    
    def get_authors(self):
        """
        Get all authors

        Returns:
            a list of dict objects representing the authors
        """
        resp = self.session.get("{}/api/authors".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def get_author_by_username(self, username):
        """
        Get author by username
            
        Args:
            username (str): The username of the author to fetch

        Returns:
            a dict object representing the author
        """
        resp = self.session.get("{}/api/authorByUsername/{}".format(self.url, username))
        resp.raise_for_status()
        return resp.json()
    
    def create_author(self, author):
        """
        Create an author

        Args:
            author (dict): A dict object representing the author to create

        Returns:
            a dict object representing the new author
        """
        def hash_password(password):
            """ Calculate the base64-encoded sha1 hash of the password for ACAS built-in authentication """
            hasher = hashlib.sha1()
            hasher.update(password.encode('utf-8'))
            return base64.b64encode(hasher.digest()).decode('utf-8')
        if 'password' in author:
            author['password'] = hash_password(author['password'])
        resp = self.session.post("{}/api/author".format(self.url),
                                    json=author)
        resp.raise_for_status()
        return resp.json()
    
    def update_author(self, author):
        """Update an author

        Args:
            author (dict): A dict object representing the author to update

        Returns:
            a dict object representing the updated author
        """
        if 'id' not in author:
            raise ValueError("id attribute of author dict is required")
        resp = self.session.put("{}/api/author/{}".format(self.url, author.get('id')),
                                    json=author)
        resp.raise_for_status()
        return resp.json()
    
    def create_authors(self, authors):
        """
        Create authors

        Args:
            authors (list): A list of dicts representing the authors to create
            
        Returns:
            a list of dict objects representing the saved authors
        """
        return [self.create_author(author) for author in authors]
    
    def update_author_roles(self, new_author_roles=None, author_roles_to_delete=None):
        """
        Update author roles
        
        Args:
            new_author_roles (list): A list of dicts representing the new author roles to create
            author_roles_to_delete (list): A list of dicts representing the author roles to delete
            
        Returns:
            a list of dict objects representing the saved author roles
        """
        body = {
            'newAuthorRoles': new_author_roles or [],
            'authorRolesToDelete': author_roles_to_delete or [],
        }
        resp = self.session.post("{}/api/updateAuthorRoles".format(self.url),
                                    json=body)
        resp.raise_for_status()
        return resp.json()
    
    def update_project_roles(self, new_author_roles=None, author_roles_to_delete=None):
        """
        Same as update author roles but with a different endpoint name.
        """
        body = {
            'newAuthorRoles': new_author_roles or [],
            'authorRolesToDelete': author_roles_to_delete or [],
        }
        resp = self.session.post("{}/api/projects/updateProjectRoles".format(self.url),
                                    json=body)
        resp.raise_for_status()
        return resp.json()
    
    def _validate_then_save_codetable(self, url_base, codeTable: dict) -> dict:
        """
        Validate a codetable and save it to the database

        Args:
            url_base (str): The base URL for the codetable
            codeTable (dict): A dict object representing the codetable to save

        Returns:
            a dict object representing the saved codetable
        """
        # Validate
        resp = self.session.post(url_base + "/validateBeforeSave", json=codeTable)
        resp.raise_for_status()
        validation_resp = resp.json()
        if type(validation_resp) is list and len(validation_resp) > 0:
            raise ValueError(validation_resp[0]['message'])
        # Create
        resp = self.session.post(url_base, json=codeTable)
        resp.raise_for_status()
        return resp.json()
    
    def get_assay_scientists(self):
        """
        Fetch the list of possible assay scientists for  assay loading
        """
        resp = self.session.get("{}/api/authors?additionalCodeType=assay&additionalCodeKind=scientist&roleName=ROLE_ACAS-USERS".format(self.url))
        resp.raise_for_status()
        return resp.json()

    def create_assay_scientist(self, code, name):
        """
        Create a new scientist for assay loading
        """
        url_base = "{}/api/codeTablesAdmin/assay/scientist".format(self.url)
        body = {'code': code, 'name': name}
        return self._validate_then_save_codetable(url_base, body)
    
    def update_assay_scientist(self, scientist: dict):
        """
        Update a scientist for  assay loading
        """
        if 'id' not in scientist:
            raise ValueError("id attribute of scientist dict is required")
        resp = self.session.put("{}/api/codeTablesAdmin/assay/scientist/{}".format(self.url, scientist['id']), json=scientist)
        resp.raise_for_status()
        return resp.json()
    
    def get_cmpdreg_scientists(self):
        """
        Fetch the list of possible lot chemists for CmpdReg
        """
        resp = self.session.get("{}/cmpdreg/scientists".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def create_cmpdreg_scientist(self, code, name):
        """
        Create a new scientist for CmpdReg
        """
        url_base = "{}/api/codeTablesAdmin/compound/scientist".format(self.url)
        body = {'code': code, 'name': name}
        return self._validate_then_save_codetable(url_base, body)
    
    def update_cmpdreg_scientist(self, scientist: dict):
        """
        Update a scientist for CmpdReg
        """
        if 'id' not in scientist:
            raise ValueError("id attribute of scientist dict is required")
        resp = self.session.put("{}/api/codeTablesAdmin/compound/scientist/{}".format(self.url, scientist['id']), json=scientist)
        resp.raise_for_status()
        return resp.json()
    
    def delete_cmpdreg_scientist(self, id: int) -> bool:
        resp = self.session.delete("{}/api/codeTablesAdmin/{}".format(self.url, id))
        resp.raise_for_status()
        return True
    
    def get_stereo_categories(self):
        """
        Get all stereo categories
        """
        resp = self.session.get("{}/api/cmpdRegAdmin/stereoCategories".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def create_stereo_category(self, code, name):
        """
        Create a new stereo category
        """
        url_base = "{}/api/cmpdRegAdmin/stereoCategories".format(self.url)
        body = {'code': code, 'name': name}
        return self._validate_then_save_codetable(url_base, body)
    
    def update_stereo_category(self, stereo_category: dict):
        """
        Update a stereo category
        """
        if 'id' not in stereo_category:
            raise ValueError("id attribute of stereo_category dict is required")
        resp = self.session.put("{}/api/cmpdRegAdmin/stereoCategories/{}".format(self.url, stereo_category['id']), json=stereo_category)
        resp.raise_for_status()
        # No return because backend doesn't return anything
    
    def delete_stereo_category(self, id: int) -> bool:
        resp = self.session.delete("{}/api/cmpdRegAdmin/stereoCategories/{}".format(self.url, id))
        resp.raise_for_status()
        return True
    
    def get_salts(self):
        """
        Get all salts
        """
        resp = self.session.get("{}/cmpdreg/salts".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def create_salt(self, abbrev, name, mol_structure):
        """
        Create a new salt
        """
        resp = self.session.post("{}/cmpdreg/salts".format(self.url), json={'abbrev': abbrev, 'name': name, 'molStructure': mol_structure})
        resp.raise_for_status()
        return resp.json()
    
    def get_physical_states(self):
        """
        Get all physical states
        """
        resp = self.session.get("{}/api/cmpdRegAdmin/physicalStates".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def create_physical_state(self, code, name):
        """
        Create a new physical state
        """
        url_base = "{}/api/cmpdRegAdmin/physicalStates".format(self.url)
        body = {'code': code, 'name': name}
        return self._validate_then_save_codetable(url_base, body)
    
    def update_physical_state(self, physical_state: dict):
        """
        Update a physical state
        """
        if 'id' not in physical_state:
            raise ValueError("id attribute of physical_state dict is required")
        resp = self.session.put("{}/api/cmpdRegAdmin/physicalStates/{}".format(self.url, physical_state['id']), json=physical_state)
        resp.raise_for_status()
        # No return because backend doesn't return anything
    
    def delete_physical_state(self, id: int) -> bool:
        resp = self.session.delete("{}/api/cmpdRegAdmin/physicalStates/{}".format(self.url, id))
        resp.raise_for_status()
        return True
    
    def get_cmpdreg_vendors(self):
        """
        Get all vendors for CmpdReg
        """
        resp = self.session.get("{}/api/cmpdRegAdmin/vendors".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def create_cmpdreg_vendor(self, code, name):
        """
        Create a new vendor for CmpdReg
        """
        url_base = "{}/api/cmpdRegAdmin/vendors".format(self.url)
        body = {'code': code, 'name': name}
        return self._validate_then_save_codetable(url_base, body)
    
    def update_cmpdreg_vendor(self, vendor: dict):
        """
        Update a vendor for CmpdReg
        """
        if 'id' not in vendor:
            raise ValueError("id attribute of vendor dict is required")
        resp = self.session.put("{}/api/cmpdRegAdmin/vendors/{}".format(self.url, vendor['id']), json=vendor)
        resp.raise_for_status()
        # No return because backend doesn't return anything
    
    def delete_cmpdreg_vendor(self, id: int) -> bool:
        resp = self.session.delete("{}/api/cmpdRegAdmin/vendors/{}".format(self.url, id))
        resp.raise_for_status()
        return True

    def setup_items(self, item_type, items):
        """Create or update items of a given typeKind 
           ACAS Admin role for this operation

        Args:
            item_type (str): Type of item to create or update
            items (list): List of items to create or update
        """
        allowed_types = ['experimenttypes', 'experimentkinds', 'statetypes', 'statekinds', 'valuetypes', 'valuekinds',
                         'labeltypes', 'labelkinds', 'ddicttypes', 'ddictkinds', 'codetables','labelsequences', 'roletypes',
                         'rolekinds', 'lsroles']
        if item_type not in allowed_types:
            raise ValueError("item_type must be one of {}".format(allowed_types))
        resp = self.session.post("{}/api/setup/{}".format(self.url, item_type), json=items)
        resp.raise_for_status()
        return resp.json()

    def get_lot_dependencies(self, lot_corp_name, include_linked_lots=True, include_analysis_group_values=True):
        """Get lot dependencies for a lot by corp name

        Args:
            lot_corp_name (str): Corp name of lot to get dependencies for
            include_linked_lots (bool): Whether to include linked lots in the response, default True.  Linked lots are purely informational as they are not a dependency preventing the lot from being deleted.
            include_analysis_group_values (bool): Whether to include analysis group values in the response, default True.  Analysis group values are purely informational as they are not a dependency preventing the lot from being deleted.

        Returns:
            A dict of the lot dependencies
            For example:
            {
                "batchCodes": [
                    "CMPD-0000001-001"
                ],
                "linkedDataExists": true,
                "linkedExperiments": [
                    {
                        "acls": {
                            "delete": true,
                            "read": true,
                            "write": true
                        },
                        "code": "EXPT-00000009",
                        "comments": "CMPD-0000001-001",
                        "description": "6 results",
                        "ignored": false,
                        "name": "BLAH",
                        "analysisGroups": [
                            {
                                "code": "AG-00000001",
                                "values": [
                                    {
                                        "id": 1,
                                        "lsKind": "key",
                                        "lsType": "numericValue",
                                        "value": 6
                                    }
                                ]
                            }
                        ],
                        "protocol": {
                            "code": "PROT-00000001",
                            "name": "Test Protocol"
                        }
                    }
                ],
                "linkedLots": [
                    {
                        "acls": {
                            "delete": false,
                            "read": true,
                            "write": true
                        },
                        "code": "CMPD-0000001-002",
                        "ignored": false,
                        "name": "CMPD-0000001-002"
                    }
                ],
                "lot": {
                    ...the lot info...
                }
            }
        Raises:
            HTTPError: If permission denied
        """

        params = {'includeLinkedLots': str(include_linked_lots).lower(), 'includeAnalysisGroupValues': str(include_analysis_group_values).lower()}
        resp = self.session.get("{}/cmpdreg/metalots/checkDependencies/corpName/{}"
                                .format(self.url, lot_corp_name),
                                params=params)
        if resp.status_code == 500:
            return None
        resp.raise_for_status()
        return resp.json()

    def delete_lot(self, lot_corp_name):
        """Delete a lot

        Args:
            lot_corp_name (str): Corp name of lot to delete

        Returns:
            A dict with "success": true if successful. For example
            {
                "success": true
            }
            Or None if there was an error
        Raises:
            HTTPError: If permission denied
        """
        resp = self.session.delete("{}/cmpdReg/metalots/corpName/{}"
                                .format(self.url, lot_corp_name))
        if resp.status_code == 500:
            return None
        resp.raise_for_status()
        return resp.json()

    def swap_parent_structures(self, corp_name1: str, corp_name2: str) -> Dict[str, str]:
        """Swap parent structures.

        Args:
            corp_name1 (str): Corporate ID of the first parent compound.
            corp_name2 (str): Corporate ID of the second parent compound.

        Returns:
            A dict with "hasError" and "errorMessage" keys. For example
            {
                "hasError": True,
                "errorMessage": "Swapping corpName1=CMPD-1 & corpName2=CMPD-2 creates duplicates."
            }
        """

        data = {'corpName1': corp_name1, 'corpName2': corp_name2}
        resp = self.session.post(f'{self.url}/cmpdreg/swapParentStructures/', json=data)
        if resp.status_code != 400:
            resp.raise_for_status()
        return resp.json()
        
    def reparent_lot(self, lot_corp_name, new_parent_corp_name, dry_run=True):
        """Reparent a lot

        Args:
            lot_corp_name (str): Corp name of lot to reparent
            new_parent_corp_name (str): Corp name of new parent
            dry_run (bool): Whether to perform a dry run, default True

        Returns:
            A dict with information about expected changes
            {
                "dependencies": {
                    "linkedDataExists": true,
                    ...other dependency data...
                },
                "modifiedBy": "bob",
                "newLot": {
                    "corpName": "CMPD-0000003-002",
                    ...other lot info...
                    "saltForm": {
                        ...salt form info...
                        "parent": {
                            "corpName": "CMPD-0000003",
                            ...other parent info...
                    },
                },
                "originalLotCorpName": "CMPD-0000001-001",
                "originalParentCorpName": "CMPD-0000001"
                "originalParentDeleted": true
            }

            Or None if there was an error
        Raises:
            HTTPError: If permission denied
        """
        data = {
            'lotCorpName': lot_corp_name,
            'parentCorpName': new_parent_corp_name
        }

        # Set dry run url param
        params = {'dryRun': str(dry_run).lower()}

        resp = self.session.post("{}/api/cmpdRegAdmin/lotServices/reparent/lot"
                                .format(self.url),
                                 params=params,
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(data))
        if resp.status_code == 500:
            return None
        resp.raise_for_status()
        return resp.json()
    
    def _get_meta_lot_by_parent_corp_name(self, parent_corp_name):
        """ Gets a MetaLot object for the first lot of the parent identified by the `parent_corp_name`
        """
        # Look up the compound to find a lot, so we can access a MetaLot
        search_results = self.cmpd_search(corpNameList=parent_corp_name)

        found_compounds = search_results['foundCompounds']

        # Return the first found_compound which matches the parent_corp_name
        found_parent = next((x for x in found_compounds if x['corpName'] == parent_corp_name), None)

        if not found_parent:
            raise RuntimeError(f'Parent corp name {parent_corp_name} could not be found')
        lot_corp_name = found_parent['lotIDs'][0]['corpName']
        return self.get_meta_lot(lot_corp_name)
    
    def get_parent_alias_kinds(self) -> List[Dict]:
        """Get the list of parent alias kinds
        """
        resp = self.session.get("{}/cmpdreg/aliases/parentAliasKinds/".format(self.url))
        resp.raise_for_status()
        return resp.json()
    
    def get_parent_aliases(self, parent_corp_name: str) -> List[str]:
        """Get the current parent aliases by parent_corp_name
        """
        meta_lot = self._get_meta_lot_by_parent_corp_name(parent_corp_name)
        # Find the parent aliases nested within the meta_lot
        alias_objects = meta_lot['lot']['saltForm']['parent']['parentAliases']
        # Flatten them to a list of strings
        aliases = [x['aliasName'] for x in alias_objects if not x['ignored']]
        return aliases
    
    def add_parent_alias(self, parent_corp_name: str, alias: str, ls_type: str = None, ls_kind: str = None) -> None:
        """Adds a new alias to the specified parent. Does not alter existing aliases.

        Args:
            parent_corp_name (str): The parent compound to add the alias to
            alias (str): The alias to add
            ls_type (str): The LS type of the alias, default None
            ls_kind (str): The LS kind of the alias, default None

        Returns:
            The updated MetaLot object
        """
        meta_lot = self._get_meta_lot_by_parent_corp_name(parent_corp_name)
        # Format the alias string into a basic object
        alias_obj = {'aliasName': alias, 'lsType': ls_type, 'lsKind': ls_kind, 'ignored': False}
        # Add it to the list of aliases
        meta_lot['lot']['saltForm']['parent']['parentAliases'].append(alias_obj)
        # Persist the change
        response = self.save_meta_lot(meta_lot)
        if len(response['errors']) != 0:
            raise RuntimeError(f'Failed to add alias {alias} to parent {parent_corp_name}: {response["errors"]}')
        return response['metalot']

    def set_parent_aliases(self, parent_corp_name: str, alias_list: List[str]) -> None:
        """Sets the aliases of the specified parent to ONLY the provided list.
        This will remove existing aliases if they are not in `alias_list`.
        To add aliases without removing existing ones, use the `add_parent_alias` method.
        """
        meta_lot = self._get_meta_lot_by_parent_corp_name(parent_corp_name)
        # Get the current aliases
        aliases = meta_lot['lot']['saltForm']['parent']['parentAliases']
        # Check existing aliasNames to see which we need to add
        existing_aliases = [x['aliasName'] for x in aliases if not x['ignored']]
        # ignore aliases not in the alias_list
        for alias_obj in aliases:
            if alias_obj['aliasName'] not in alias_list:
                # Ignore aliases not in the new list
                alias_obj['ignored'] = True
        # Add new ones
        for alias in alias_list:
            if alias not in existing_aliases:
                new_alias_obj = {'aliasName': alias}
                aliases.append(new_alias_obj)
        
        # Persist the update
        self.save_meta_lot(meta_lot)
    
    def edit_parent(self, parent, dry_run=True) -> Tuple[bool, Dict]:
        """Makes changes to an existing parent.
        Returns (status, data) where `status` is a bool representing success (True) or failure (False)
        `data` is a dict of supporting information. In case of failure it contains a list of duplicates
        In case of success it contains a list of affected lots
        """
        # Validate
        resp = self.session.post("{}/cmpdreg/validateParent".format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(parent))
        resp.raise_for_status()
        resp_data = resp.json()
        success = True
        # Check whether the response just has a list of affected lots or whether
        # It has a list of dupeParents
        if 'parentUnique' in resp_data and not resp_data['parentUnique']:
            success = False
        # If just doing validation, or if validation failed, return response
        if dry_run or not success:
            return success, resp.json()
        # Otherwise continue to updateParent
        resp = self.session.post("{}/cmpdreg/updateParent".format(self.url),
                                 headers={'Content-Type': "application/json"},
                                 data=json.dumps(parent))
        resp.raise_for_status()
        resp_data = resp.json()
        return success, resp_data



#!/usr/bin/env python

"""Tests for `acasclient` package."""

import os
from datetime import datetime
import uuid
import logging
from pathlib import Path

from acasclient.ddict import ACASDDict, ACASLsThingDDict
from acasclient.lsthing import (BlobValue, CodeValue, FileValue, LsThingValue,
                                SimpleLsThing, get_lsKind_to_lsvalue, datetime_to_ts)
from acasclient.validation import ValidationResult, get_validation_response
from tests.test_acasclient import BaseAcasClientTest


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# Constants
from tests.project_thing import (
    PROJECT_METADATA, PROJECT, STATUS, PROJECT_STATUS,PROCEDURE_DOCUMENT,PDF_DOCUMENT, 
    NAME_KEY, IS_RESTRICTED_KEY, STATUS_KEY, START_DATE_KEY, DESCRIPTION_KEY, PROJECT_NAME,
    START_DATE, PDF_DOCUMENT_KEY, PROCEDURE_DOCUMENT_KEY, PARENT_PROJECT_KEY, ACTIVE, INACTIVE,
    PROJECT_NUMBER_KEY, PROJECT_NUMBER, Project
)

FWD_ITX = 'relates to'
BACK_ITX = 'is related to'

class TestLsThing(BaseAcasClientTest):
    """Tests for `acasclient lsthing` package model."""

    # See test_acasclient.BaseAcasClientTest for setUp

    def tearDown(self):
        """Tear down test fixtures, if any."""
        files_to_delete = ['dummy.pdf', 'dummy2.PDF']
        for f in files_to_delete:
            file = Path(f)
            if file.exists():
                os.remove(file)
        super().tearDown()

    # Helpers
    def _get_path(self, file_name):
        path = Path(__file__).resolve().parent\
            .joinpath('test_acasclient', file_name)
        return path

    def _get_bytes(self, file_path):
        with open(file_path, "rb") as in_file:
            file_bytes = in_file.read()
        return file_bytes

    def _check_blob_equal(self, blob_value, orig_file_name, orig_bytes):
        self.assertEqual(blob_value.comments, orig_file_name)
        data = blob_value.download_data(self.client)
        self.assertEqual(data, orig_bytes)

    def _check_file(self, file_path, orig_file_name, orig_file_path):
        # Check file names match
        self.assertEqual(Path(file_path).name, orig_file_name)
        # Check file contents match
        new_bytes = self._get_bytes(file_path)
        orig_bytes = self._get_bytes(orig_file_path)
        self.assertEqual(new_bytes, orig_bytes)

    def _test_codevalue_missing_error(self, message, value, code_type, code_kind, code_origin):
        base_msg = "'{code}' is not yet in the database as a valid '{code_kind}'. Please double-check the spelling and correct your data if you expect this to match an existing term. If this is a novel valid term, please contact your administrator to add it to the following dictionary: Code Type: {code_type}, Code Kind: {code_kind}, Code Origin: {code_origin}"
        expected_msg = base_msg.format(code=value, code_type=code_type, code_kind=code_kind, code_origin=code_origin)
        self.assertEqual(message, expected_msg)


    # Tests
    def test_000_simple_ls_thing_save(self):
        """Test saving simple ls thing."""
        name = str(uuid.uuid4())
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now()
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        self.assertIsNone(newProject.code_name)
        self.assertIsNone(newProject._ls_thing.id)
        newProject.save(self.client)
        self.assertIsNotNone(newProject.code_name)
        self.assertIsNotNone(newProject._ls_thing.id)

    def test_001_simple_ls_thing_save_with_blob_value(self):
        """Test saving simple ls thing with blob value."""
        name = str(uuid.uuid4())
        file_name = 'blob_test.png'
        blob_test_path = Path(__file__).resolve().parent\
            .joinpath('test_acasclient', file_name)

        # Get the file bytes for testing
        in_file = open(blob_test_path, "rb")
        file_bytes = in_file.read()
        in_file.close()

        # Save with Path path
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PROCEDURE_DOCUMENT_KEY: blob_test_path
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.save(self.client)
        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)

        # Save with string path
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PROCEDURE_DOCUMENT_KEY: str(blob_test_path)
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.save(self.client)
        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)

        # Write to a file by providing a full file path
        custom_file_name = "my.png"
        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(full_file_path=Path(self.tempdir, custom_file_name))
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.name, custom_file_name)

        # Write to a file by providing a full file path as a string
        custom_file_name = "my.png"
        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(full_file_path=str(Path(self.tempdir, custom_file_name)))
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.name, custom_file_name)

        # Write to a file by providing a folder
        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path=self.tempdir)
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.name, file_name)

        # Write to a file by providing a folder as a string
        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path=str(self.tempdir))
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.name, file_name)

        # Write to a file by providing a folder and custom file name
        output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path=self.tempdir, file_name=custom_file_name)
        self.assertTrue(output_file.exists())
        self.assertEqual(output_file.name, custom_file_name)

        # Write to a bad folder path fails gracefully
        with self.assertRaises(ValueError):
            output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path="GARBAGE")
        try:
            output_file = newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT].write_to_file(folder_path="GARBAGE")
        except ValueError as err:
            self.assertIn("does not exist", err.args[0])

        # Make sure a bad path fails gracefully
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PROCEDURE_DOCUMENT_KEY: "SOMEGARBAGEPATH"
        }
        with self.assertRaises(ValueError):
            newProject = Project(recorded_by=self.client.username, **meta_dict)
        try:
            newProject = Project(recorded_by=self.client.username, **meta_dict)
        except ValueError as err:
            self.assertIn("does not exist", err.args[0])

        # Make sure passing a directory fails gracefully
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PROCEDURE_DOCUMENT_KEY: self.tempdir
        }
        with self.assertRaises(ValueError):
            newProject = Project(recorded_by=self.client.username, **meta_dict)
        try:
            newProject = Project(recorded_by=self.client.username, **meta_dict)
        except ValueError as err:
            self.assertIn("not a file", err.args[0])

    def test_002_update_blob_value(self):
        """Test saving simple ls thing with blob value, then updating the blobValue."""

        # Create a project with first blobValue
        name = str(uuid.uuid4())
        file_name = 'blob_test.png'
        file_path = self._get_path(file_name)
        file_bytes = self._get_bytes(file_path)

        # Save with Path path
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PROCEDURE_DOCUMENT_KEY: file_path
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.save(self.client)
        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)

        # Then update with a different file
        file_name = '1_1_Generic.xlsx'
        file_path = self._get_path(file_name)
        file_bytes = self._get_bytes(file_path)
        new_blob_val = BlobValue(file_path=file_path)
        newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT] = new_blob_val
        newProject.save(self.client)
        self._check_blob_equal(newProject.metadata[PROJECT_METADATA][PROCEDURE_DOCUMENT], file_name, file_bytes)

    def test_003_simple_ls_thing_save_with_file_value(self):
        """Test saving simple ls thing with file value."""
        name = str(uuid.uuid4())
        file_name = 'dummy.pdf'
        file_test_path = Path(__file__).resolve().parent\
            .joinpath('test_acasclient', file_name)
        file_name_2 = 'dummy2.PDF'
        file_test_path_2 = Path(__file__).resolve().parent\
            .joinpath('test_acasclient', file_name_2)

        # Save with Path value
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PDF_DOCUMENT_KEY: file_test_path
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.save(self.client)
        # Write file locally and compare
        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
        downloaded_path = fv.download_to_disk(self.client)
        self._check_file(downloaded_path, file_name, file_test_path)

        # Save with string value
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
            PDF_DOCUMENT_KEY: str(file_test_path)
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.save(self.client)
        # Write file locally and compare
        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
        downloaded_path = fv.download_to_disk(self.client)
        self._check_file(downloaded_path, file_name, file_test_path)

        # Save with Path file_path
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT] = FileValue(file_path=file_test_path)
        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].value)
        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].comments)
        newProject.save(self.client)
        # Write file locally and compare
        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
        downloaded_path = fv.download_to_disk(self.client)
        self._check_file(downloaded_path, file_name, file_test_path)

        # Save with string file_path
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now(),
        }
        newProject = Project(recorded_by=self.client.username, **meta_dict)
        newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT] = FileValue(file_path=str(file_test_path))
        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].value)
        self.assertIsNotNone(newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT].comments)
        newProject.save(self.client)
        # Write file locally and compare
        fv = newProject.metadata[PROJECT_METADATA][PDF_DOCUMENT]
        downloaded_path = fv.download_to_disk(self.client)
        self._check_file(downloaded_path, file_name, file_test_path)

        # Write to a bad folder path fails gracefully
        with self.assertRaises(ValueError):
            fv.download_to_disk(self.client, folder_path="GARBAGE")
        try:
            fv.download_to_disk(self.client, folder_path="GARBAGE")
        except ValueError as err:
            self.assertIn("does not exist", err.args[0])

        # Test updating other values on a saved Thing
        saved_project = Project.get_by_code(newProject.code_name, self.client, Project.ls_type, Project.ls_kind)
        STATUS_DDICT = ACASDDict(PROJECT, STATUS)
        saved_project.metadata[PROJECT_METADATA][PROJECT_STATUS] = CodeValue(INACTIVE, ddict=STATUS_DDICT)
        saved_project.save(self.client)

        # Test updating FileValue on a saved Thing
        saved_project = Project.get_by_code(newProject.code_name, self.client, Project.ls_type, Project.ls_kind)
        saved_project.metadata[PROJECT_METADATA][PDF_DOCUMENT] = FileValue(file_path=file_test_path_2)
        saved_project.save(self.client)
        fv = saved_project.metadata[PROJECT_METADATA][PDF_DOCUMENT]
        downloaded_path = fv.download_to_disk(self.client)
        self._check_file(downloaded_path, file_name_2, file_test_path_2)

    def test_004_get_lskind_to_ls_values(self):
        """
        Verify `get_lskind_to_lsvalue` adds the `unit_kind` if present to the
        `lskind` value.
        """

        # No unit_kind in LsThingValue
        lsthing_value = LsThingValue(
            ls_type='foo',
            ls_kind='bar',
            numeric_value=4.5,
        )
        lskind_to_lsvalue = get_lsKind_to_lsvalue([lsthing_value])
        assert len(lskind_to_lsvalue) == 1
        assert 'bar' in lskind_to_lsvalue

        # No unit_kind in LsThingValue
        lsthing_value = LsThingValue(
            ls_type='foo',
            ls_kind='bar',
            numeric_value=4.5,
            unit_kind='baz')
        lskind_to_lsvalue = get_lsKind_to_lsvalue([lsthing_value])
        assert len(lskind_to_lsvalue) == 1
        assert 'bar (baz)' in lskind_to_lsvalue

    def test_005_create_interactions(self):
        name = str(uuid.uuid4())
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now()
        }
        name_2 = str(uuid.uuid4())
        meta_dict_2 = {
            NAME_KEY: name_2,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: ACTIVE,
            START_DATE_KEY: datetime.now()
        }
        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
        proj_1.save(self.client)
        proj_2 = Project(recorded_by=self.client.username, **meta_dict_2)
        proj_2.save(self.client)
        # add an interaction
        proj_1.add_link(FWD_ITX, proj_2, recorded_by=self.client.username)
        assert len(proj_1.links) == 1
        # save the interaction
        proj_1.save(self.client)
        # Fetch project 1 and look at the interaction
        fresh_proj_1 = Project.get_by_code(proj_1.code_name, self.client, Project.ls_type, Project.ls_kind)
        assert len(fresh_proj_1.links) == 1
        assert len(fresh_proj_1._ls_thing.second_ls_things) == 1
        assert fresh_proj_1._ls_thing.second_ls_things[0].ls_kind == f'{Project.ls_type}_{Project.ls_type}'
        # check if save populated second_ls_things properly
        # FIXME: save is not properly populating second_ls_things
        # assert len(proj_1._ls_thing.second_ls_things) == 1
        itx_ls_thing_ls_thing = fresh_proj_1._ls_thing.second_ls_things[0]
        assert itx_ls_thing_ls_thing.id is not None
        # Fetch project 2 again and look at the interaction in reverse
        fresh_proj_2 = Project.get_by_code(proj_2.code_name, self.client, Project.ls_type, Project.ls_kind)
        assert len(fresh_proj_2.links) == 1
        back_itx = fresh_proj_2.links[0]
        assert back_itx.verb == BACK_ITX
        # Run advanced search by interaction
        # Forward interaction query
        second_itx_listings = [
            {
                "interactionType": FWD_ITX,
                "thingType": Project.ls_type,
                "thingKind": Project.ls_kind,
                "operator": "=",
                "thingCodeName": proj_2.code_name
            }]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       second_itx_listings=second_itx_listings,
                                       codes_only=True,
                                       max_results=1000,
                                       combine_terms_with_and=True)
        assert len(results) == 1
        assert results[0] == proj_1.code_name
        # Backwards interaction query
        first_itx_listings = [
            {
                "interactionType": FWD_ITX,
                "thingType": Project.ls_type,
                "thingKind": Project.ls_kind,
                "operator": "=",
                "thingCodeName": proj_1.code_name
            }]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       first_itx_listings=first_itx_listings,
                                       codes_only=True,
                                       max_results=1000,
                                       combine_terms_with_and=True)
        assert len(results) == 1
        assert results[0] == proj_2.code_name

        # Save new ls things to test interaction subject and object type customization
        name_3 = str(uuid.uuid4())
        name_4 = str(uuid.uuid4())
        meta_dict.update({'name': name_3})
        meta_dict_2.update({'name': name_4})
        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
        proj_1.save(self.client)
        proj_2 = Project(recorded_by=self.client.username, **meta_dict_2)
        proj_2.save(self.client)
        # add an interaction and override the subject and object types
        subject_type = 'test'
        object_type = 'me'
        proj_1.add_link(FWD_ITX, proj_2, recorded_by=self.client.username, subject_type=subject_type, object_type=object_type)
        assert len(proj_1.links) == 1
        # save the interaction
        proj_1.save(self.client)
        # Fetch project 1 and look at the interaction
        fresh_proj_1 = Project.get_by_code(proj_1.code_name, self.client, Project.ls_type, Project.ls_kind)
        assert len(fresh_proj_1.links) == 1
        assert len(fresh_proj_1._ls_thing.second_ls_things) == 1
        assert fresh_proj_1._ls_thing.second_ls_things[0].ls_kind == f'{subject_type}_{object_type}'

    def test_006_simple_thing_overide_label_and_state_types(self):
        class ExampleThing(SimpleLsThing):
            ls_type = "parent"
            ls_kind = "Example Thing"
            ID_LS_TYPE = "corpName"
            NAME_LS_TYPE = "MyNameType"
            ALIAS_LS_TYPE = "MyAliasType"
            METADATA_LS_TYPE = "MyMetaDataType"
            RESULTS_LS_TYPE = "MyResultsType"

            def __init__(self, name=None, alias=None, id=None, recorded_by=None, metadata={}, results={}, ls_thing=None):
                # ID "corpName" "Example Thing" will be created on save because
                # it cooresponds to a saved label sequence with matching type and kind attributes
                # Its important to send in a '' on initial save
                ids = {'Example Thing': ''}
                names = {'MyNameKind': name}
                aliases = {'MyAliasKind': alias}

                super().__init__(ls_type=self.ls_type, ls_kind=self.ls_kind, names=names, aliases=aliases, ids=ids, recorded_by=recorded_by,
                                 metadata=metadata, results=results, ls_thing=ls_thing)

        name = str(uuid.uuid4())
        alias = str(uuid.uuid4())
        meta_dict = {
            'alias': alias,
            'name': name,
            'results': {
                'experimental': {
                    'My Result': 134,
                    'My Result Date': datetime.now()
                }
            },
            'metadata': {
                'general': {
                    'Species': "Rat",
                    'Description': "This is an in vitro pharamacology assay"
                }
            }
        }
        newExampleThing = ExampleThing(recorded_by=self.client.username, **meta_dict)
        newExampleThing.save(self.client)
        # Confirm that the label types are being saved correctly
        assert len(newExampleThing._ls_thing.ls_labels) == 3
        assert newExampleThing._ls_thing.ls_labels[0].ls_type in [ExampleThing.ID_LS_TYPE, ExampleThing.NAME_LS_TYPE, ExampleThing.ALIAS_LS_TYPE]
        assert newExampleThing._ls_thing.ls_labels[1].ls_type in [ExampleThing.ID_LS_TYPE, ExampleThing.NAME_LS_TYPE, ExampleThing.ALIAS_LS_TYPE]
        assert newExampleThing._ls_thing.ls_labels[2].ls_type in [ExampleThing.ID_LS_TYPE, ExampleThing.NAME_LS_TYPE, ExampleThing.ALIAS_LS_TYPE]

        # Confirm that the state types are being saved correctly
        assert newExampleThing._ls_thing.ls_states[0].ls_type in [ExampleThing.METADATA_LS_TYPE, ExampleThing.RESULTS_LS_TYPE]
        assert newExampleThing._ls_thing.ls_states[1].ls_type in [ExampleThing.METADATA_LS_TYPE, ExampleThing.RESULTS_LS_TYPE]

        # Verify that the example thing picked up a corpName label sequence
        fresh_example_thing = ExampleThing.get_by_code(newExampleThing.code_name, self.client, ExampleThing.ls_type, ExampleThing.ls_kind)
        # The label sequence for example thing is in the format ET-000001 so check it fetched a new label and is in the ids field
        assert fresh_example_thing.ids['Example Thing'].startswith('ET-')

    def test_004_get_lskind_to_ls_values(self):
        """
        Verify `get_lskind_to_lsvalue` adds the `unit_kind` if present to the
        `lskind` value.
        """

        # No unit_kind in LsThingValue
        lsthing_value = LsThingValue(
            ls_type='foo',
            ls_kind='bar',
            numeric_value=4.5,
        )
        lskind_to_lsvalue = get_lsKind_to_lsvalue([lsthing_value])
        assert len(lskind_to_lsvalue) == 1
        assert 'bar' in lskind_to_lsvalue

        # No unit_kind in LsThingValue
        lsthing_value = LsThingValue(
            ls_type='foo',
            ls_kind='bar',
            numeric_value=4.5,
            unit_kind='baz')
        lskind_to_lsvalue = get_lsKind_to_lsvalue([lsthing_value])
        assert len(lskind_to_lsvalue) == 1
        assert 'bar (baz)' in lskind_to_lsvalue

    def test_007_advanced_search_interactions(self):

        # Create project 1
        name = str(uuid.uuid4())
        status_1 = str(uuid.uuid4())
        desc_1 = str(uuid.uuid4())
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: status_1,
            START_DATE_KEY: datetime.now(),
            DESCRIPTION_KEY: desc_1
        }

        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
        # skip CodeValue validation since this is not a valid status
        proj_1.save(self.client, skip_validation=True)

        # Create project 2
        name_2 = str(uuid.uuid4())
        status_2 = str(uuid.uuid4())
        desc_2 = str(uuid.uuid4())
        meta_dict_2 = {
            NAME_KEY: name_2,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: status_2,
            START_DATE_KEY: datetime.now(),
            DESCRIPTION_KEY: desc_2
        }
        proj_2 = Project(recorded_by=self.client.username, **meta_dict_2)
        # skip CodeValue validation since this is not a valid status
        proj_2.save(self.client, skip_validation=True)

        # Add interactions between projects
        proj_1.add_link(FWD_ITX, proj_2, recorded_by=self.client.username)
        assert len(proj_1.links) == 1
        # skip CodeValue validation since this is not a valid status
        proj_1.save(self.client, skip_validation=True)

        # Run advanced search by interaction w/value matching on the interaction thing
        # Forward interaction query w/interaction thing values
        # Code value search
        second_itx_listings = [
            {
                "lsType": FWD_ITX,
                "lsKind": "project_project",
                "thingType": Project.ls_type,
                "thingKind": Project.ls_kind,
                "thingValues": [
                    {
                    	"value": status_2,
						"stateType": "metadata",
						"stateKind": "project metadata",
						"valueType": "codeValue",
						"valueKind": "project status",
						"operator": "~"
                    }
                ]
            }
        ]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       second_itx_listings=second_itx_listings,
                                       format="nestedfull",
                                       max_results=1000,
                                       combine_terms_with_and=True)
        assert len(results) == 1
        assert results[0]["codeName"] == proj_1.code_name

        # String value search
        second_itx_listings = [
            {
                "interactionType": FWD_ITX,
                "thingType": Project.ls_type,
                "thingKind": Project.ls_kind,
                "thingCodeName": proj_2.code_name,
                "thingValues": [
                    {
                    	"value": desc_2,
						"stateType": "metadata",
						"stateKind": "project metadata",
						"valueType": "stringValue",
						"valueKind": "description",
						"operator": "~"
                    }
                ]
            }
        ]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       second_itx_listings=second_itx_listings,
                                       format="nestedfull",
                                       max_results=1000,
                                       combine_terms_with_and=True)
        assert len(results) == 1
        assert results[0]["codeName"] == proj_1.code_name

        # Upper case search using ~ which should still match
        second_itx_listings = [
            {
                "interactionType": FWD_ITX,
                "thingType": Project.ls_type,
                "thingKind": Project.ls_kind,
                "thingCodeName": proj_2.code_name,
                "thingValues": [
                    {
                    	"value": desc_2.upper(),
						"stateType": "metadata",
						"stateKind": "project metadata",
						"valueType": "stringValue",
						"valueKind": "description",
						"operator": "~"
                    }
                ]
            }
        ]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       second_itx_listings=second_itx_listings,
                                       format="nestedfull",
                                       max_results=1000,
                                       combine_terms_with_and=True)
        assert len(results) == 1
        assert results[0]["codeName"] == proj_1.code_name

        # String value search make sure it doesn't return the wrong project (this should not return any results)
        second_itx_listings = [
            {
                "interactionType": FWD_ITX,
                "thingType": Project.ls_type,
                "thingKind": Project.ls_kind,
                "thingCodeName": proj_2.code_name,
                "thingValues": [
                    {
                    	"value": desc_1,
						"stateType": "metadata",
						"stateKind": "project metadata",
						"valueType": "stringValue",
						"valueKind": "description",
						"operator": "~"
                    }
                ]
            }
        ]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       second_itx_listings=second_itx_listings,
                                       format="nestedfull",
                                       max_results=1000,
                                       combine_terms_with_and=True)
        assert len(results) == 0

    def test_008_validate_codevalue(self):
        """Test creating CodeValues using code + code_type + code_kind + code_origin
        Confirm that invalid code values are rejected by validate method.
        """
        # Create project 1
        name = str(uuid.uuid4())
        status_1 = str(uuid.uuid4())
        desc_1 = str(uuid.uuid4())
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: status_1,
            START_DATE_KEY: datetime.now(),
            DESCRIPTION_KEY: desc_1
        }
        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
        valid = proj_1.validate(self.client)
        assert not valid
        messages = valid.get_messages()
        assert len(messages) == 1
        self._test_codevalue_missing_error(messages[0], status_1, PROJECT, STATUS, 'ACAS DDict')
        # Now test timing of one-by-one validation with 20 projects versus doing it in bulk
        # Create 20 valid projects
        meta_dict[STATUS_KEY] = ACTIVE
        projects = [Project(recorded_by=self.client.username, **meta_dict) for i in range(20)]
        single_start = datetime.now()
        for proj in projects:
            valid = proj.validate(self.client)
            assert valid
        single_end = datetime.now()
        single_duration = single_end - single_start
        logger.info(f"Single validation took {single_duration}")
        valid = Project.validate_list(self.client, projects)
        assert valid
        bulk_end = datetime.now()
        bulk_duration = bulk_end - single_end
        logger.info(f"Bulk validation took {bulk_duration}")
        assert single_duration > bulk_duration

    def test_009_validate_ddicts(self):
        """Test creating CodeValues using code + DDict
        Confirm that invalid code values are rejected by validate method.
        """
        # Create project 1
        name = str(uuid.uuid4())
        desc_1 = str(uuid.uuid4())
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            START_DATE_KEY: datetime.now(),
            DESCRIPTION_KEY: desc_1
        }

        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
        # set status to a CodeValue constructed with a DDict
        STATUS_DDICT = ACASDDict(PROJECT, STATUS)
        proj_1.metadata[PROJECT_METADATA][PROJECT_STATUS] = CodeValue(ACTIVE, ddict=STATUS_DDICT)
        # Because we are referencing a valid status, this should not raise an error
        valid = proj_1.validate(self.client)
        assert valid
        assert len(valid.get_messages()) == 0
        # Now try setting status to an invalid CodeValue
        status_1 = str(uuid.uuid4())
        proj_1.metadata[PROJECT_METADATA][PROJECT_STATUS] = CodeValue(status_1, ddict=STATUS_DDICT)
        valid = proj_1.validate(self.client)
        assert not valid
        assert len(valid.get_messages()) == 1
        self._test_codevalue_missing_error(valid.get_messages()[0], status_1, PROJECT, STATUS, 'ACAS DDict')
        # Now try adding a CodeValue that references an LsThing
        # First we create and save a new LsThing so we can get a code_name
        proj_1.metadata[PROJECT_METADATA][PROJECT_STATUS] = CodeValue(ACTIVE, ddict=STATUS_DDICT)
        proj_1.save(self.client)
        # Then we create a new project 2 and set PARENT_PROJECT_KEY to reference `proj_1`
        name_2 = str(uuid.uuid4())
        desc_2 = str(uuid.uuid4())
        PARENT_PROJECT_DDICT = ACASLsThingDDict(PROJECT, PROJECT)
        meta_dict = {
            NAME_KEY: name_2,
            IS_RESTRICTED_KEY: True,
            START_DATE_KEY: datetime.now(),
            DESCRIPTION_KEY: desc_2
        }
        proj_2 = Project(recorded_by=self.client.username, **meta_dict)
        proj_2.metadata[PROJECT_METADATA][PARENT_PROJECT_KEY] = CodeValue(proj_1.code_name, ddict=PARENT_PROJECT_DDICT)
        # Because we are referencing a valid LsThing, this should be valid
        valid = proj_2.validate(self.client)
        assert valid
        # Now try setting parent project to an invalid CodeValue and confirm validation fails
        bad_project_code = str(uuid.uuid4())
        proj_2.metadata[PROJECT_METADATA][PARENT_PROJECT_KEY] = CodeValue(bad_project_code, ddict=PARENT_PROJECT_DDICT)
        valid = proj_2.validate(self.client)
        assert not valid
        self._test_codevalue_missing_error(valid.get_messages()[0], bad_project_code, PROJECT, PROJECT, 'ACAS LsThing')
        # Generate validation response and check it
        validation_response = get_validation_response(valid, ls_thing=proj_2)
        # Confirm the structure of the response body
        assert validation_response.get("commit") is False
        self.assertEqual(validation_response.get("transaction_id"), -1)
        assert validation_response.get("hasError") is True
        assert validation_response.get("hasWarning") is False
        assert validation_response.get("results") is not None
        assert validation_response.get("results").get("thing") is not None
        # Check we have one message and it is an error
        self.assertEqual(len(validation_response.get("errorMessages")), 1)
        msg = validation_response.get("errorMessages")[0]
        self.assertEqual(msg.get("errorLevel"), "error")
        # Do a loose check of the html summary and confirm it contains mention of our bad code
        assert bad_project_code in validation_response['results']['htmlSummary']


    def test_008_get_by_code(self):
        """
        If no lsthing entry is found for the given `code_name`, `ls_type` and
        `ls_kind` then `get_by_code` should raise KeyError.
        """

        with self.assertRaises(KeyError):
            _ = SimpleLsThing.get_by_code(code_name='baz', client=self.client,
            ls_type='foo', ls_kind='bar')


    def test_009_advanced_search_flat_response(self):

        # Create project 1
        name = str(uuid.uuid4())
        status_1 = str(uuid.uuid4())
        desc_1 = str(uuid.uuid4())
        start_date = datetime.now()
        project_number = 1234
        meta_dict = {
            NAME_KEY: name,
            IS_RESTRICTED_KEY: True,
            STATUS_KEY: status_1,
            START_DATE_KEY: start_date,
            DESCRIPTION_KEY: desc_1,
            PROJECT_NUMBER_KEY: project_number
        }

        proj_1 = Project(recorded_by=self.client.username, **meta_dict)
        # skip CodeValue validation since this is not a valid status
        proj_1.save(self.client, skip_validation=True)

        # Run advanced search by searching by description which is unique
        value_listings = [
            {
                "value": desc_1,
                "stateType": "metadata",
                "stateKind": "project metadata",
                "valueType": "stringValue",
                "valueKind": DESCRIPTION_KEY,
                "operator": "="
            }
        ]

        # Return listing values
        return_status_key = "My Status"
        return_description_key = "My Description"
        return_startdate_key = "My Start Date"
        return_name_key = "My Name"
        return_project_number_key = "My Project Number"
        return_listings = [
            {
                # Code value return listing
                "key": return_status_key,
                "stateType": "metadata",
                "stateKind": "project metadata",
                "valueType": "codeValue",
                "valueKind": PROJECT_STATUS
            },
            {
                # String value return listing
                "key": return_description_key,
                "stateType": "metadata",
                "stateKind": "project metadata",
                "valueType": "stringValue",
                "valueKind": DESCRIPTION_KEY
            },
            {
                # Date value return listing
                "key": return_startdate_key,
                "stateType": "metadata",
                "stateKind": "project metadata",
                "valueType": "dateValue",
                "valueKind": START_DATE
            },
            {
                # Label return listing
                "key": return_name_key,
                "labelType": "name",
                "labelKind": PROJECT_NAME,
            },
            {
                # Numeric Value return listing
                "key": return_project_number_key,
                "stateType": "metadata",
                "stateKind": "project metadata",
                "valueType": "numericValue",
                "valueKind": PROJECT_NUMBER
            }
        ]
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       value_listings=value_listings,
                                       format="flat",
                                       max_results=1000,
                                       combine_terms_with_and=True,
                                       return_listings={"thingValues": return_listings})
        assert len(results) == 1

        # Verify code name returns
        assert results[0]["codeName"] == proj_1.code_name

        # Verify code value returns
        assert return_status_key in results[0]
        assert results[0][return_status_key] == status_1

        # Verify string value returns
        assert return_description_key in results[0]
        assert results[0][return_description_key] == desc_1

        # Verify date value returns
        assert return_startdate_key in results[0]
        assert results[0][return_startdate_key] == datetime_to_ts(start_date)

        # Verify numeric value returns
        assert return_project_number_key in results[0]
        assert results[0][return_project_number_key] == project_number

        # Verify label returns
        assert return_name_key in results[0]
        assert results[0][return_name_key] == name

        # Without thing attributes listings, it should return all thing attributes
        # assert that all thing attributes are returned
        thingAttributes = {"id", "codeName", "lsType", "lsKind", "recordedBy", "recordedDate"}
        for thingAttribute in thingAttributes:
            assert thingAttribute in results[0]

        # Verify that when we specify thing attributes listings, only those thing attributes are returned
        return_thing_attributes = {"id", "codeName"}
        results = self.client\
            .advanced_search_ls_things('project', 'project', None,
                                       value_listings=value_listings,
                                       format="flat",
                                       max_results=1000,
                                       combine_terms_with_and=True,
                                       return_listings={"thingAttributes": list(return_thing_attributes), "thingValues": return_listings})

        shouldNotExistAttributes = thingAttributes.symmetric_difference(return_thing_attributes)
        for thingAttribute in shouldNotExistAttributes:
            assert thingAttribute not in results[0]
        for thingAttribute in return_thing_attributes:
            assert thingAttribute in results[0]
 
        # Verify that adding a duplicate key gives an error by passing the same return listing twice in the same request
        with self.assertRaises(ValueError) as context:
            self.client\
                .advanced_search_ls_things('project', 'project', None,
                                        value_listings=value_listings,
                                        format="flat",
                                        max_results=1000,
                                        combine_terms_with_and=True,
                                        return_listings={"thingValues": return_listings + return_listings})
        assert 'duplicate' in str(context.exception)

        # Verify that reserved word "codeName" cannot be used as a return listing key
        bad_request_code_name_reserved = return_listings
        bad_request_code_name_reserved[0]["key"] = "codeName"
        with self.assertRaises(ValueError) as context:
            self.client\
                .advanced_search_ls_things('project', 'project', None,
                                        value_listings=value_listings,
                                        format="flat",
                                        max_results=1000,
                                        combine_terms_with_and=True,
                                        return_listings={"thingValues": bad_request_code_name_reserved})
        assert 'returned as a thing attribute' in str(context.exception)


class TestBlobValue(BaseAcasClientTest):

    def test_as_dict(self):
        """
        Verify `as_dict` returns the instance attributes mapped to their value.
        """

        value, comments, id = [65, 67], "blob test", None
        blob_value = BlobValue(value=value, comments=comments, id=id)
        blob_value_dict = blob_value.as_dict()

        assert len(blob_value_dict) == len(BlobValue._fields)
        assert blob_value_dict['value'] == value
        assert blob_value_dict['comments'] == comments
        assert blob_value_dict['id'] == id

class TestValidationResponse(BaseAcasClientTest):

    def test_001_response_with_errors_and_warnings(self):
        """
        Test creating a ValidationResponse with errors and warnings.
        """
        ERR_MSG = 'error 1'
        WARN_MSG = 'warning 1'
        SUMM_MSG = 'summary 1'
        # Create a warning
        warn = ValidationResult(True, [WARN_MSG])
        assert warn
        assert len(warn.get_messages()) == 1
        assert len(warn.get_warnings()) == 1
        assert warn.get_warnings()[0] == WARN_MSG
        assert len(warn.get_errors()) == 0

        # Create an error
        err = ValidationResult(False, [ERR_MSG])
        assert not err
        assert len(err.get_messages()) == 1
        assert len(err.get_warnings()) == 0
        assert len(err.get_errors()) == 1
        assert err.get_errors()[0] == ERR_MSG

        # Create a summary
        summary = ValidationResult(True, summaries=[SUMM_MSG])

        # Add them
        valid = warn + err
        assert not valid
        assert len(valid.get_messages()) == 2
        assert valid.get_messages()[0] == ERR_MSG
        assert valid.get_messages()[1] == WARN_MSG
        assert len(valid.get_errors()) == 1
        assert valid.get_errors()[0] == ERR_MSG
        assert len(valid.get_warnings()) == 1
        assert valid.get_warnings()[0] == WARN_MSG

        # Add in the summary
        valid = valid + summary
        assert not valid
        assert len(valid.get_messages()) == 3
        assert valid.get_messages()[2] == SUMM_MSG
        assert len(valid.get_summaries()) == 1
        assert valid.get_summaries()[0] == SUMM_MSG

        # Generate and check a response
        response = get_validation_response(valid)
        # Confirm the structure of the response body
        assert response.get("commit") is False
        self.assertEqual(response.get("transaction_id"), -1)
        assert response.get("hasError") is True
        assert response.get("hasWarning") is True
        assert response.get("results") is not None
        # Check we have two messages: one error, one warning
        self.assertEqual(len(response.get("errorMessages")), 2)
        self.assertEqual(response.get("errorMessages")[0].get("errorLevel"), "error")
        self.assertEqual(response.get("errorMessages")[1].get("errorLevel"), "warning")
        # Confirm HTML summary contains both messages
        assert ERR_MSG in response['results']['htmlSummary']
        assert WARN_MSG in response['results']['htmlSummary']

        # Create a ValidationResult with error and warning at once
        valid = ValidationResult(False, errors=[ERR_MSG], warnings=[WARN_MSG])
        assert len(valid.get_messages()) == 2

        # Create an invalid ValidationResult by giving an error message but saying it is valid
        try:
            invalid = ValidationResult(True, errors=[ERR_MSG], warnings=[WARN_MSG])
            assert False
        except ValueError as e:
            assert str(e) == "ValidationResult cannot be valid and contain error messages"

    def test_002_html_summary(self):
        """
        Test the HTML Summary generated by ValidationResult
        """
        ERR_INSTRUCT = "<p>Please fix the following errors and use the 'Back' button at the bottom of this screen to upload a new version of the data.</p>"
        WARN_INSTRUCT = "<p>Please review the warnings and summary before uploading.</p>"
        SUCCESS_INSTRUCT = "<p>Please review the summary before uploading.</p>"
        COMMIT_INSTRUCT = "<p>Upload completed.</p>"
        ERR_1 = 'error 1'
        ERR_2 = 'error 2'
        WARN_1 = 'warning 1'
        WARN_2 = 'warning 2'
        SUMM_1 = 'summary 1'
        SUMM_2 = 'summary 2'
        # No errors or warnings
        res = ValidationResult(True)
        html = get_validation_response(res).get('results').get('htmlSummary')
        self.assertEqual(html.strip(), SUCCESS_INSTRUCT)
        # 1 errors 1 warning
        res = ValidationResult(False, errors=[ERR_1], warnings=[WARN_1])
        html = get_validation_response(res).get('results').get('htmlSummary')
        self.assertIn(ERR_INSTRUCT, html)
        self.assertIn('<h4 style="color:red">Errors: 1 </h4>', html)
        self.assertIn('<h4>Warnings: 1 </h4>', html)
        self.assertIn(f"<li>{ERR_1}</li>", html)
        self.assertIn(f"<li>{WARN_1}</li>", html)
        # Multiple errors & warnings, including dupes
        res = ValidationResult(False, errors=[ERR_1, ERR_2, ERR_1], warnings=[WARN_1, WARN_1, WARN_2])
        html = get_validation_response(res).get('results').get('htmlSummary')
        self.assertIn('<h4 style="color:red">Errors: 2 </h4>', html)
        self.assertIn('<h4>Warnings: 2 </h4>', html)
        self.assertIn(f'<li>2 occurrences of: {ERR_1}</li>', html)
        self.assertNotIn(f"<li>{ERR_1}</li>", html)
        self.assertIn(f"<li>{ERR_2}</li>", html)
        self.assertIn(f'<li>2 occurrences of: {WARN_1}</li>', html)
        self.assertIn(f"<li>{WARN_2}</li>", html)
        # Warnings, no errors
        res = ValidationResult(True, warnings=[WARN_1, WARN_2])
        html = get_validation_response(res).get('results').get('htmlSummary')
        self.assertIn(WARN_INSTRUCT, html)
        self.assertIn('<h4>Warnings: 2 </h4>', html)
        self.assertNotIn('<h4 style="color:red">Errors:', html)
        # Errors, no warnings
        res = ValidationResult(False, errors=[ERR_1, ERR_2])
        html = get_validation_response(res).get('results').get('htmlSummary')
        self.assertIn(ERR_INSTRUCT, html)
        self.assertIn('<h4 style="color:red">Errors: 2 </h4>', html)
        self.assertNotIn('<h4>Warnings:', html)
        # Errors, warnings, and summaries
        res = ValidationResult(False, errors=[ERR_1, ERR_2], warnings=[WARN_1, WARN_2], summaries=[SUMM_1, SUMM_2])
        html = get_validation_response(res).get('results').get('htmlSummary')
        self.assertIn(ERR_INSTRUCT, html)
        self.assertIn('<h4 style="color:red">Errors: 2 </h4>', html)
        self.assertIn('<h4>Warnings: 2 </h4>', html)
        self.assertIn('<h4>Summary</h4>', html)
        self.assertIn(f"<li>{SUMM_1}</li>", html)
        self.assertIn(f"<li>{SUMM_2}</li>", html)
        # Commit should hide warnings and show a different message
        res = ValidationResult(True, warnings=[WARN_1], summaries=[SUMM_1])
        html = get_validation_response(res, commit=True).get('results').get('htmlSummary')
        self.assertIn(COMMIT_INSTRUCT, html)
        self.assertNotIn(ERR_INSTRUCT, html)
        self.assertNotIn('<h4 style="color:red">Errors:', html)
        self.assertNotIn('<h4>Warnings:', html)
        self.assertIn('<h4>Summary</h4>', html)
        self.assertIn(f"<li>{SUMM_1}</li>", html)



class StereoCategory:
	def __init__(self, id: str, name: str, code: str, version: str) -> None:
		self.id = id
		self.name = name
		self.code = code
		self.version = version

	def get_id(self) -> int:
		return self.id

	def get_name(self) -> str:
		return self.name

	def get_code(self) -> str:
		return self.code

	def get_version(self) -> str:
		return self.version

	# Define as dict for json serialization
	def as_dict(self) -> dict:
		return {
			'id': self.id,
			'name': self.name,
			'code': self.code,
			'version': self.version
		}

	def __str__(self) -> str:
		return "id: " + id + " , " + "name: " + name + " , " + "code: " + code + " , " + "version: " + version



import decorator
from collections import defaultdict
from typing import List

# Author: Anand Kumar

################################################################################
# Validation Result
################################################################################


class ValidationResult(object):
    """
    Class stores the validation results data. Each validation result stores
    an optional list of errors and warnings.
    Usage::

            messages = ['Invalid file type', 'Multiple structures found']
            val_res = ValidationResult(is_valid=True, messages=messages)

            if not val_res:
                print('\n'.join(val_res.get_messages()))

    """

    def __init__(self, is_valid: bool, messages: List[str] = None, errors: List[str] = None, warnings: List[str] = None, summaries: List[str] = None):
        self._is_valid = is_valid
        self._errors = errors or []
        self._warnings = warnings or []
        self._summaries = summaries or []
        # For simple messages, classify them as errors or warnings based on the validity of the result
        if messages:
            if is_valid:
                self._warnings += messages
            else:
                self._errors += messages
        # Check we are not in the invalid state of "valid" with errors
        if is_valid and self._errors:
            raise ValueError("ValidationResult cannot be valid and contain error messages")

    def __bool__(self) -> bool:
        return self._is_valid

    def __str__(self) -> str:
        res = 'VALID' if self._is_valid else 'INVALID'
        if self._errors:
            res += '\nERRORS:\n' + '\n'.join(self._errors)
        if self._warnings:
            res += '\nWARNINGS:\n' + '\n'.join(self._warnings)
        if self._summaries:
            res += '\nSUMMARIES:\n' + '\n'.join(self._summaries)
        return res

    def __repr__(self) -> str:
        return str(repr)

    def __add__(self, other):
        is_valid = self._is_valid and other._is_valid
        errors = self._errors + other._errors
        warnings = self._warnings + other._warnings
        summaries = self._summaries + other._summaries
        return ValidationResult(is_valid=is_valid, errors=errors, warnings=warnings, summaries=summaries)

    def get_messages(self) -> List[str]:
        return self._errors + self._warnings + self._summaries
    
    def get_errors(self) -> List[str]:
        return self._errors
    
    def get_warnings(self) -> List[str]:
        return self._warnings
    
    def get_summaries(self) -> List[str]:
        return self._summaries


@decorator.decorator
def validation_result(func, *args, **kwargs):
    """
    Encapsulate `func` returned values inside `ValidationResult`.

    :return: Validation result.
    :rtype: ValidationResult
    :raises TypeError: If `func` doesn't return a bool or a tuple or
            ValidationResult.
    """

    result = func(*args, **kwargs)

    if result is None:
        result = True

    if isinstance(result, bool):
        return ValidationResult(is_valid=result)
    elif isinstance(result, tuple):
        is_valid = result[0]
        messages = []
        if len(result) > 1:
            _messages = result[1]
            if isinstance(_messages, list):
                messages = _messages
            elif isinstance(_messages, str):
                messages = [_messages]
            else:
                raise TypeError("Expected a tuple of a boolean and "
                                f"list/str, got '{_messages}'")
        return ValidationResult(is_valid=is_valid, messages=messages)
    elif isinstance(result, ValidationResult):
        return result

    raise TypeError(f"Expected bool or tuple return values, got '{result}'")


def _dedupe_messages(messages: List[str]) -> List[str]:
    """
    Deduplicate messages and count occurrences.

    :param messages: list of messages
    :type messages: list

    :return: list of deduplicated messages
    :rtype: list
    """
    messages_dict = defaultdict(int)
    for msg in messages:
        messages_dict[msg] += 1
    deduped_messages = []
    for msg, count in messages_dict.items():
        if count > 1:
            deduped_messages.append(f'{count} occurrences of: {msg}')
        else:
            deduped_messages.append(msg)
    return deduped_messages

def get_validation_response(validation_result, ls_thing=None, commit=False, transaction_id=-1) -> dict:
    """
    :param validation_result: validation result object
    :type validation_result: validations.ValidationResult

    :param ls_thing: ls_thing of an entity
    :type ls_thing: dict

    :param commit: if the data was committed to the database
    :type commit: bool

    :param transaction_id: id of the transaction
    :type transaction_id: int

    :return: validation response for the given result
    :rtype: dict
    """
    errors = validation_result.get_errors()
    warnings = validation_result.get_warnings()
    summaries = validation_result.get_summaries()
    
    has_errors = len(errors) > 0
    has_warnings = len(warnings) > 0
    
    # Deduplicate error and warning messages
    errors = _dedupe_messages(errors)
    warnings = _dedupe_messages(warnings)

    # Build error_messages list of dicts
    error_messages = []
    for msg in errors:
        error_messages.append({'message': msg, 'errorLevel': 'error'})
    for msg in warnings:
        error_messages.append({'message': msg, 'errorLevel': 'warning'})
    # Format HTML
    html_summary = _get_html_summary(errors, warnings, summaries, commit)
    resp_dict = {
        'commit': commit,
        'transaction_id': transaction_id,
        'results': {
            'htmlSummary': html_summary,
        },
        'hasError': has_errors,
        'hasWarning': has_warnings,
        'errorMessages': error_messages,
    }
    if ls_thing:
        resp_dict['results']['thing'] = ls_thing
    return resp_dict


def _get_html_summary(errors, warnings, summaries, commit) -> str:
    """
    Format HTML summary for the validation result.
    """
    # Lay out HTML template structure
    HTML_SUMMARY_TEMPLATE = """<p>{instructions}</p>
    {errors_block}
    {warnings_block}
    {summary_block}
    """
    ERRORS_TEMPLATE = """<h4 style=\"color:red\">Errors: {count} </h4>
    <ul>{message_list}</ul>
    """
    WARNINGS_TEMPLATE = """<h4>Warnings: {count} </h4>
    <p>Warnings provide information on issues found in the uploaded data. You can proceed with warnings; however, it is recommended that, if possible, you make the changes suggested by the warnings and upload a new version of the data by using the 'Back' button at the bottom of this screen.<p>
    <ul>{message_list}</ul>"""
    SUMMARY_TEMPLATE = """<h4>Summary</h4>
    <p>Information:</p>
    <ul>{message_list}</ul>"""
    MESSAGE_TEMPLATE = """<li>{message}</li>"""

    # Set up variables
    instructions = ''
    errors_block = ''
    warnings_block = ''
    summary_block = ''
    if errors:
        error_messages = '\n'.join(
            [MESSAGE_TEMPLATE.format(message=error) for error in errors])
        errors_block = ERRORS_TEMPLATE.format(
            count=len(errors), message_list=error_messages)
    if warnings:
        warnings_messages = '\n'.join(
            [MESSAGE_TEMPLATE.format(message=warning) for warning in warnings])
        warnings_block = WARNINGS_TEMPLATE.format(
            count=len(warnings), message_list=warnings_messages)
    if summaries:
        summary_messages = '\n'.join(
            [MESSAGE_TEMPLATE.format(message=summary) for summary in summaries])
        summary_block = SUMMARY_TEMPLATE.format(message_list=summary_messages)
    if commit:
        instructions = 'Upload completed.'
        # hide warnings and errors if we've already committed
        errors_block = ''
        warnings_block = ''
    elif errors:
        instructions = "Please fix the following errors and use the 'Back' button at the bottom of this screen to upload a new version of the data."
    elif warnings:
        instructions = "Please review the warnings and summary before uploading."
    else:
        instructions = "Please review the summary before uploading."
    # Format HTML
    html_summary = HTML_SUMMARY_TEMPLATE.format(
        instructions=instructions,
        errors_block=errors_block,
        warnings_block=warnings_block,
        summary_block=summary_block)
    return html_summary


from __future__ import annotations
from .ddict import ACASDDict
from enum import Enum
import types
import logging
from .acasclient import client

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class AdditionalScientistType(Enum):
    """Enum for additional scientist types."""
    COMPOUND = ACASDDict('compound', 'scientist')
    ASSAY = ACASDDict('assay', 'scientist')


class AdditionalScientist():
    """Additional Scientist class."""
    def __init__(self, type: AdditionalScientistType, id: int = None, code: str = None, name: str = None, ignored: bool = None):
        self.id = id
        self.ignored = ignored
        self.code = code
        self.name = name
        self.type = type

    def save(self, client: client) -> AdditionalScientist:
        """Save the scientist to the server."""
        if self.type == AdditionalScientistType.COMPOUND:
            resp = client.create_cmpdreg_scientist(self.code, self.name)
        elif self.type == AdditionalScientistType.ASSAY:
            resp = client.create_assay_scientist(self.code, self.name)
        self.id = resp['id']
        return self

    def as_dict(self) -> dict:
        return {
            'id': self.id,
            'code': self.code,
            'name': self.name,
            'ignored': self.ignored,
            'type': self.type.name
        }


class AdditionalCompoundScientist(AdditionalScientist):
    """Additional Compound Scientist class."""

    def __init__(self, id: int = None, code: str = None, name: str = None, ignored: bool = None):
        super().__init__(AdditionalScientistType.COMPOUND, id, code, name, ignored)


class AdditionalAssayScientist(AdditionalScientist):
    """Additional Assay Scientist class."""

    def __init__(self, id: int = None, code: str = None, name: str = None, ignored: bool = None):
        super().__init__(AdditionalScientistType.ASSAY, id, code, name, ignored)


def meta_lots_to_dict_array(meta_lots: list[dict]) -> list[dict]:
    """Converts a list of meta lots to a list of flattened dictionaries which specific fields which are suitable for upload via CmpdReg BulkLoader."""
    return [meta_lot_to_dict(meta_lot) for meta_lot in meta_lots]


def meta_lot_to_dict(meta_lot: dict) -> dict:
    """Converts a meta lot to a dictionary into a flat dictionary of specific fields."""
    parent_common_names = [parent_alias["aliasName"] for parent_alias in meta_lot["lot"]["saltForm"]["parent"]["parentAliases"] if parent_alias["lsKind"] == "Common Name"]

    # Check for salts
    salt_abrevs = [iso_salt['salt']["abbrev"] for iso_salt in meta_lot["isosalts"]]
    salt_equivs = [str(iso_salt["equivalents"]) for iso_salt in meta_lot["isosalts"]]

    # Synthesis date is in the format MM/DD/YYYY
    # However Creg isn't good at converting the date back to the correct format
    # So we need to convert it to YYYY-MM-DD
    synthesis_date = meta_lot["lot"]["synthesisDate"]
    if synthesis_date is not None:
        synthesis_date = synthesis_date[6:] + '-' + synthesis_date[:2] + '-' + synthesis_date[3:5]

    # If lot amount is filled in, then we need to fill in lot barcode with the lot corp name
    lot_barcode = meta_lot["lot"]["corpName"] if meta_lot["lot"]["amount"] is not None else None

    return_dict = {
        'mol': meta_lot["lot"]['asDrawnStruct'] if meta_lot["lot"]['asDrawnStruct'] is not None else meta_lot["lot"]["parent"]["molStructure"],
        "id": meta_lot["lot"]["id"],
        "name": meta_lot["lot"]["corpName"],
        "parent_common_name": '; '.join(parent_common_names) if len(parent_common_names) > 0 else None,
        "parent_corp_name": meta_lot["lot"]["parent"]["corpName"],
        "lot_amount":  meta_lot["lot"]["amount"],
        "lot_amount_units":  meta_lot["lot"]["amountUnits"]["code"] if meta_lot["lot"]["amountUnits"] is not None else None,
        "lot_color": meta_lot["lot"]["color"],
        "lot_synthesis_date": synthesis_date,
        "lot_notebook_page": meta_lot["lot"]["notebookPage"],
        "lot_corp_name": meta_lot["lot"]["corpName"],
        "lot_number": meta_lot["lot"]["lotNumber"],
        "lot_purity": meta_lot["lot"]["purity"],
        "lot_purity_opertator": meta_lot["lot"]["purityOperator"]["code"] if meta_lot["lot"]["purityOperator"] is not None else None,
        "lot_comments": meta_lot["lot"]["comments"],
        "lot_chemist": meta_lot["lot"]["chemist"],
        "lot_solution_amount":  meta_lot["lot"]["solutionAmount"],
        "lot_solution_amount_units": meta_lot["lot"]["solutionAmountUnits"]["code"] if meta_lot["lot"]["solutionAmountUnits"] is not None else None,
        "lot_supplier": meta_lot["lot"]["supplier"],
        "lot_supplier_id": meta_lot["lot"]["supplierID"],
        "project": meta_lot["lot"]["project"],
        "parent_stereo_category": meta_lot["lot"]["parent"]["stereoCategory"]['code'],
        "parent_stereo_comment": meta_lot["lot"]["parent"]["stereoComment"],
        "lot_is_virtual": meta_lot["lot"]["isVirtual"],
        "lot_supplier_lot": meta_lot["lot"]["supplierLot"],
        "lot_salt_abbrev": '; '.join(salt_abrevs) if len(salt_abrevs) > 0 else None,
        "lot_salt_equivalents": '; '.join(salt_equivs) if len(salt_equivs) > 0 else None,
        "lot_barcode": lot_barcode
    }
    return return_dict


def convert_dict_to_type(dict: dict, type_name: str) -> object:
    """Converts a dictionary to a dynamic python type and instaniates a new object of that type.
    
    :param dict: The dictionary to convert
    :param type_name: The name of the new type
    :return: A new type with the keys of the dictionary as attributes
    """

    # Add a _fields attribute with the keys of the dictionary
    dict['_fields'] = list(dict.keys())

    # A method added to the dynamic type which can be used to conver the type back to a dictionary
    def as_dict(self) -> dict:
        """
        Return a map of attribute name and attribute values stored on the
        instance.
        Note: Only attributes stored in `FileValue._fields` will be returned.
        """
        return {
            field: getattr(self, field, None)
            for field in self._fields
        }

    # Create a new type with the dictionary as the attributes
    # and instantiates a new object of that type
    new_type = type(type_name, (object,), dict)()

    # Add the as_dict method to the new type
    new_type.as_dict = types.MethodType(as_dict, new_type)
    return new_type


def convert_object_to_sdf(object) -> str:
    """Converts an object to an SDF with each of the keys and values as mol properties.
    
    :param object: The object to convert which must have an as_dict method and a mol attribute
    :return: The object as an SDF
    """

    # Convert object to dictionary
    object_dict = object.as_dict()

    # Get the molfile
    molfile = object_dict['mol']

    # Remove the molfile from the dictionary
    del object_dict['mol']

    # Change the dict keys from snake case to space case
    object_dict = {key.replace('_', ' ').title(): value for key, value in object_dict.items()}

    # Strip any '$'s and newlines from the end of the molfile
    molfile = molfile.rstrip().rstrip('$').rstrip()

    # Format of SDF is:
    # {molfile}
    # >  <{key}>
    # {value}
    #
    # >  <{key}>
    # {value}
    #
    # $$$$
    # So we need to add the >  <{key}> to the beginning of each value
    object_dict = {f'>  <{key}>': value for key, value in object_dict.items()}

    # Join the object keys and values like the format above
    sdf = '\n'.join([f'{key}\n{value}\n' for key, value in object_dict.items() if value is not None and value != ''])

    # Add the molfile to the beginning of the sdf and add the $$$$ at the end
    sdf = f'{molfile}\n{sdf}\n$$$$'

    return sdf


def create_cmpd_reg_mapping_from_object(object) -> list[dict]:
    """Creates a compound registration mapping array of dictionaries from an object.
    
    Example input:
        {
            "lot_color": "Blue",
            "lot_corp_name": "Test Corp",
            "lot_number": "1",
            "parent_stereo_category": "R",
            "parent_stereo_comment": ,
            "lot_chemist": "Test Chemist"
        }
    Example output:
        [
            {
                "dbProperty": "Lot Color",
                "defaultVal": null,
                "required": false,
                "sdfProperty": "Lot Color"
            },
            {
                "dbProperty": "Lot Corp Name",
                "defaultVal": null,
                "required": false,
                "sdfProperty": "Lot Corp Name"
            },
            {
                "dbProperty": "Lot Number",
                "defaultVal": null,
                "required": false,
                "sdfProperty": "Lot Number"
            },
            {
                "dbProperty": "Parent Stereo Category",
                "defaultVal": null,
                "required": false,
                "sdfProperty": "Parent Stereo Category"
            },
            {
                "dbProperty": "Parent Stereo Comment",
                "defaultVal": null,
                "required": false,
                "sdfProperty": "Parent Stereo Comment"
            },
            {
                "dbProperty": "Lot Chemist",
                "defaultVal": null,
                "required": false,
                "sdfProperty": "Lot Chemist"
            }
        ]
    :param object: The object to convert which must have an as_dict method
    :return: The mapping array of dictionaries which creates a compound registration mapping
    """
    # Convert object to dictionary
    object_dict = object.as_dict()

    # For each of the keys in the object, reate an object of the form:
    # "dbProperty": "Lot Amount Units",
    # "defaultVal": null,
    # "required": false,
    # "sdfProperty": "Lot Amount Units"

    # Get the keys of the object
    keys = object_dict.keys()

    # Create the mapping
    mapping = []
    for key in keys:
        mapping.append({
            "dbProperty": key.replace('_', ' ').title(),
            "defaultVal": None,
            "required": False,
            "sdfProperty": key.replace('_', ' ').title()
        })

    return mapping
